from docker.models.nodes import Node
from docker.models.images import Image
from docker.models.services import Service
from docker.models.secrets import Secret
from docker.types import RestartPolicy, EndpointSpec, SecretReference, NetworkAttachmentConfig

from .api_common import PapdlAPIContext, get_papdl_service, get_papdl_secret
from .common import AppType
from ..slice.slice import Slice
from typing import List, TypedDict, Tuple, Dict
from os import path, mkdir, PathLike
from random import randint
from OpenSSL import crypto
from copy import deepcopy
import sys
from pathlib import Path


class PapdlRegistryAPI:
    def __init__(self, context: PapdlAPIContext):
        self.context = context

    def cert_gen(self, cert_path: PathLike, key_path: PathLike):
        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, 4096)
        cert = crypto.X509()
        cert.get_subject().C = self.cert_specs["C"]
        cert.get_subject().ST = self.cert_specs["ST"]
        cert.get_subject().L = self.cert_specs["L"]
        cert.get_subject().O = self.cert_specs["O"]
        cert.get_subject().OU = self.cert_specs["OU"]
        cert.get_subject().CN = self.cert_specs["CN"]
        cert.get_subject().emailAddress = self.cert_specs["emailAddress"]
        cert.set_serial_number(randint(1, sys.maxsize - 1))
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(self.cert_specs["validity_seconds"])
        cert.set_issuer(cert.get_subject())
        cert.set_pubkey(k)
        cert.sign(k, 'sha256')

        with open(cert_path, "wt") as f:
            f.write(
                crypto.dump_certificate(
                    crypto.FILETYPE_PEM,
                    cert).decode("utf-8"))

        with open(key_path, "wt") as f:
            f.write(
                crypto.dump_privatekey(
                    crypto.FILETYPE_PEM,
                    k).decode("utf-8"))

    def secret_gen(self) -> Tuple[Secret, Secret]:
        cert_path = path.join(
            self.context.dircontext["api_module_path"],
            "certificates",
            "registry.crt")
        key_path = path.join(
            self.context.dircontext["api_module_path"],
            "certificates",
            "registry.key")
        if (not (Path(cert_path).is_file() and Path(key_path).is_file())):
            self.cert_gen(cert_path, key_path)

        registry_docker_cert = get_papdl_secret(
            self.context, labels={
                "type": "cert"}, name="registry.crt")
        registry_docker_key = get_papdl_secret(
            self.context, labels={
                "type": "key"}, name="registry.key")

        registry_crt: Secret = None
        if (len(registry_docker_cert) == 0):
            with open(cert_path, "rb") as f:
                cert_data = f.read()
                registry_crt = self.context.client.secrets.create(
                    name="registry.crt", data=cert_data, labels={
                        "papdl": "true", "type": "cert"})
        else:
            registry_crt = registry_docker_cert[0]

        registry_key: Secret = None
        if (len(registry_docker_key) == 0):
            with open(key_path, "rb") as f:
                key_data = f.read()
                registry_key = self.context.client.secrets.create(
                    name="registry.key", data=key_data, labels={
                        "papdl": "true", "type": "key"})
        else:
            registry_key = registry_docker_key[0]

        return registry_crt, registry_key

    def assign_registry_node(self):
        manager_node = self.context.client.nodes.list(
            filters={'role': "manager"})[0]
        manager_node_spec = deepcopy(manager_node.attrs["Spec"])
        manager_node_spec["Labels"]["registry"] = 'true'
        manager_node.update(manager_node_spec)

    def spawn_registry(self) -> Service:
        self.assign_registry_node()
        registry_services = get_papdl_service(
            self.context, labels={"type": AppType.REGISTRY.value})
        if (len(registry_services) != 0):
            return registry_services[0]

        self.context.logger.info(f"Spawning Registry service...")
        self.context.client.images.pull("registry", tag="latest")
        crt, key = self.secret_gen()
        registry_volume_path = path.join(
            self.context.dircontext["api_module_path"],
            "registry_volume")
        endpoint_spec = EndpointSpec(ports={443: 443})

        sr_crt = SecretReference(secret_id=crt.id, secret_name=crt.name)
        sr_key = SecretReference(secret_id=key.id, secret_name=key.name)
        rp = RestartPolicy(condition="any")
        nac = NetworkAttachmentConfig(self.context.network.name)

        service = self.context.client.services.create(
            image="registry",
            name="registry",
            constraints=[f"node.labels.registry==true"],
            mounts=[f"{registry_volume_path}:/var/lib/registry"],
            env=[
                "REGISTRY_HTTP_ADDR=0.0.0.0:443",
                "REGISTRY_HTTP_TLS_CERTIFICATE=/run/secrets/registry.crt",
                "REGISTRY_HTTP_TLS_KEY=/run/secrets/registry.key",
            ],
            secrets=[sr_crt, sr_key],
            maxreplicas=1,
            endpoint_spec=endpoint_spec,
            networks=[nac],
            labels={
                "papdl": "true",
                "type": AppType.REGISTRY.value
            },
            restart_policy=rp
        )
        return service



import click
from ..backend.api_common import PapdlAPIContext, get_papdl_service, get_papdl_image, get_papdl_network, get_papdl_secret
from ..backend.common import AppType


@click.option("-b", "--benchmark_images", show_default=True,
              is_flag=True, flag_value=True)
@click.option("-B", "--benchmark_services", show_default=True,
              is_flag=True, flag_value=True)
@click.option("-s", "--slice_images", show_default=True,
              is_flag=True, flag_value=True)
@click.option("-S", "--slice_services", show_default=True,
              is_flag=True, flag_value=True)
@click.option("-o", "--orchestrator_images", show_default=True,
              is_flag=True, flag_value=True)
@click.option("-O", "--orchestrator_services", show_default=True,
              is_flag=True, flag_value=True)
@click.option("-R", "--registry_service", show_default=True,
              is_flag=True, flag_value=True)
@click.option("-I", "--iperf_services", show_default=True,
              is_flag=True, flag_value=True)
@click.option("-n", "--network", show_default=True,
              is_flag=True, flag_value=True)
@click.option("-x", "--secrets", show_default=True,
              is_flag=True, flag_value=True)
@click.command()
def clean(
    benchmark_images: bool,
    benchmark_services: bool,
    slice_images: bool,
    slice_services: bool,
    orchestrator_images:bool,
    orchestrator_services:bool,
    registry_service: bool,
    iperf_services: bool,
    network: bool,
    secrets: bool,

):
    pac = PapdlAPIContext()

    if (benchmark_images):
        [i.remove(force=True)
         for i in get_papdl_image(pac, labels={"type": AppType.BENCHMARKER.value})]
    if (benchmark_services):
        [s.remove() for s in get_papdl_service(
            pac, labels={"type": AppType.BENCHMARKER.value})]

    if (slice_images):
        [i.remove(force=True)
         for i in get_papdl_image(pac, labels={"type": AppType.SLICE.value})]
    if (slice_services):
        [s.remove() for s in get_papdl_service(pac, labels={"type": AppType.SLICE.value})]
        
    if (orchestrator_images):
        [i.remove(force=True)
         for i in get_papdl_image(pac, labels={"type": AppType.ORCHESTRATOR.value})]
    if (orchestrator_services):
        [s.remove() for s in get_papdl_service(pac, labels={"type": AppType.ORCHESTRATOR.value})]
    

    if (registry_service):
        [s.remove() for s in get_papdl_service(
            pac, labels={"type": "registry"})]
    if (iperf_services):
        [s.remove() for s in get_papdl_service(pac, labels={"type": "iperf"})]
    if (network):
        [n.remove() for n in get_papdl_network(pac)]
    if (secrets):
        [s.remove() for s in get_papdl_secret(pac)]



from time import time
import numpy as np
from getpass import getuser
from glob import glob
from keras.models import Model, load_model
from typing import Dict, TypedDict, List
import json
from io import BytesIO
from os import stat, environ, path
import iperf3
from pythonping import ping
from getpass import getuser
import tensorflow as tf
import psutil
import gc
from multiprocessing import Pool
import contextlib
import uuid
from pympler.asizeof import asizeof

class Config(TypedDict):
    model_test_number_of_repeats: int
    model_test_batch_size: int
    bandwidth_test_duration_sec: int
    latency_test_count: int
    free_memory_multiplier:float

config: Config

INPUT_MULTIPLIER = 6


def load_benchmark_configs()->Config:
    global config
    config = Config(
        model_test_batch_size=int(environ.get("MODEL_TEST_BATCH_SIZE")),
        model_test_number_of_repeats=int(environ.get("MODEL_TEST_NUMBER_OF_REPEATS")),
        bandwidth_test_duration_sec=int(environ.get("BANDWIDTH_TEST_DURATION_SEC")),
        latency_test_count=int(environ.get("LATENCY_TEST_COUNT")),
        free_memory_multiplier=float(environ.get("FREE_MEMORY_MULTIPLIER"))
    )

def isDebug()->bool:
    return int(environ.get("DEBUG")) == 1



def get_available_memory():
    return psutil.virtual_memory().free

def model_memory_usage(model, *, batch_size:int):
    default_dtype = tf.keras.backend.floatx()
    shapes_mem_count = 0
    internal_model_mem_count = 0
    for layer in model.layers:
        if isinstance(layer, tf.keras.Model):
            internal_model_mem_count += model_memory_usage(
                layer, batch_size=batch_size
            )
        single_layer_mem = tf.as_dtype(layer.dtype or default_dtype).size
        out_shape = layer.output_shape
        if isinstance(out_shape, list):
            out_shape = out_shape[0]
        for s in out_shape:
            if s is None:
                continue
            single_layer_mem *= s
        shapes_mem_count += single_layer_mem 

    trainable_count = sum(
        [tf.keras.backend.count_params(p) for p in model.trainable_weights]
    )
    non_trainable_count = sum(
        [tf.keras.backend.count_params(p) for p in model.non_trainable_weights]
    )
    return trainable_count + non_trainable_count

def hidden_layer_input_usage(model, *, batch_size: int):
    default_dtype = tf.keras.backend.floatx()
    shapes_mem_count = 0
    internal_model_mem_count = 0
    for layer in model.layers:
        if isinstance(layer, tf.keras.Model):
            internal_model_mem_count += hidden_layer_input_usage(
                layer, batch_size=batch_size
            )
        single_layer_mem = tf.as_dtype(layer.dtype or default_dtype).size
        out_shape = layer.output_shape
        if isinstance(out_shape, list):
            out_shape = out_shape[0]
        for s in out_shape:
            if s is None:
                continue
            single_layer_mem *= s
        shapes_mem_count += single_layer_mem
    
    return (
        batch_size * shapes_mem_count
        + internal_model_mem_count
    )

def get_input_memory_multiplier(dimensions):
    return np.prod(np.array(dimensions)) * INPUT_MULTIPLIER

def total_memory_usage(model,dimensions, *, batch_size:int):
    return model_memory_usage(model,batch_size=batch_size) + hidden_layer_input_usage(model,batch_size=batch_size) + get_input_memory_multiplier(dimensions)

def load_network_benchmark_ips() -> List[str]:
    target_networks = environ.get("PAPDL_WORKERS").split(" ")
    print(target_networks,flush=True)
    return target_networks

def get_available_memory():
    return psutil.virtual_memory().free


user_folder = glob(f"/home/*/")[0]
model_paths = glob(f"{user_folder}models/*/")
papdl_workers = load_network_benchmark_ips()

def benchmark_network(papdl_workers:List[str]) -> Dict:
    global config
    result = {}
    for ip in papdl_workers:

        # BANDWIDTH TEST
        client = iperf3.Client()
        client.duration = config["bandwidth_test_duration_sec"]
        client.server_hostname = ip
        client.port = 5201
        r_iperf: iperf3.TestResult = client.run()
        r_ping = ping(ip, count=config["latency_test_count"])

        result[ip] = {
            "bandwidth": {
                "sent_bps": r_iperf.sent_bps,
                "received_bps": r_iperf.received_bps
            },
            "latency": {
                "rtt_min_ms": r_ping.rtt_min_ms,
                "rtt_avg_ms": r_ping.rtt_avg_ms,
                "rtt_max_ms": r_ping.rtt_max_ms
            }
        }
        print(result[ip],flush=True)
        del client
    return result

    
def single_model_benchmark(args:Dict):
    mp = args["mp"]
    fmm = args["fmm"]
    mtbs = args["mtbs"]
    mtnor = args["mtnor"]
    
    free_memory = get_available_memory()
    model:tf.keras.Model = load_model(mp)
    model_name = model.name
    dimensions = (mtbs,) + model.input_shape[1:]
    total = total_memory_usage(model=model,dimensions=dimensions, batch_size=mtbs)
    print(f"Loaded model: {model_name} from path: {mp} with input_dims: {dimensions}",flush=True)
    
    if(total > free_memory * fmm):
        print(f"Skipping benchmarking as memory threshold {total} has met {free_memory}.",flush=True)

        result = {}

        result["benchmark_size"] = float("inf")
        result["benchmark_time"] = float("inf")

        result["benchmark_model_memory_usage"] = model_memory_usage(model,batch_size=mtbs)
        result["benchmark_hidden_and_input_memory_usage"] = hidden_layer_input_usage(model,batch_size=mtbs)
        result["benchmark_input_memory_multiplier"] = get_input_memory_multiplier(dimensions)
        result["model_name"] = model_name

        del model
        tf.compat.v1.reset_default_graph()
        gc.collect()
        return result
        
    print(f"Running benchmarking as memory threshold {free_memory} has not been met for mode {total}")
    sample_input = np.random.random_sample(dimensions)
    start = time()
    for i in range(mtnor):
        tmp_out = model(sample_input,training=False)
        del tmp_out
        gc.collect()
    end = time()
    output:np.ndarray = model(sample_input,training=False)
    size = asizeof(output)

    result = {}

    result["benchmark_size"] = size
    result["benchmark_time"] = (end - start) / mtbs

    result["benchmark_model_memory_usage"] = model_memory_usage(model,batch_size=mtbs)
    result["benchmark_hidden_and_input_memory_usage"] = hidden_layer_input_usage(model,batch_size=mtbs)
    result["benchmark_input_memory_multiplier"] = get_input_memory_multiplier(dimensions)

    result["model_name"] = model_name

    del model
    del output
    del sample_input
    tf.compat.v1.reset_default_graph()
    gc.collect()
    return result
    


def benchmark_models(model_paths=model_paths):
    global config
    result = {}
    with contextlib.closing(Pool(1)) as po:
        pool_args = []
        for mp in model_paths:
            pool_args.append(
                {
                    "mp":mp, 
                    "fmm":config["free_memory_multiplier"], 
                    "mtbs": config["model_test_batch_size"],
                    "mtnor": config["model_test_number_of_repeats"]
                }
            )

        pool_result = po.map_async(single_model_benchmark, pool_args)
        completed_pool_results = pool_result.get()
        for r in completed_pool_results:
            result[r["model_name"]] = {}

            result[r["model_name"]]["benchmark_size"] = r["benchmark_size"]

            result[r["model_name"]]["benchmark_time"] = r["benchmark_time"]

            result[r["model_name"]]["benchmark_model_memory_usage"] = r["benchmark_model_memory_usage"]

            result[r["model_name"]]["benchmark_hidden_and_input_memory_usage"] = r["benchmark_hidden_and_input_memory_usage"]

            result[r["model_name"]]["benchmark_input_memory_multiplier"] = r["benchmark_input_memory_multiplier"]
    return result
        

load_benchmark_configs()
print(f"Loaded config: {config}",flush=True)
benchmark_result = {"free_memory": get_available_memory()}
benchmark_result["network_performance"] = benchmark_network(papdl_workers=papdl_workers)
benchmark_result["model_performance"] = benchmark_models(model_paths=model_paths)

def convert_np(o):
    if isinstance(o, np.int64): return int(o)  
    raise TypeError
print("[BENCHMARK]" + json.dumps(benchmark_result,default=convert_np),flush=True)


from collections.abc import Iterable
from enum import Enum
from typing import List
from keras import models,layers


class Slice:
    def __init__(self):
        self.model: models.Model = None
        self.input_layer = 0
        self.output_layer = 0
        self.second_prediction = 0
        self.output_size = 0

# Recursively gets the output of a layer, used to build up a submodel


def get_output_of_layer(layer, new_input, starting_layer_name):
    global layer_outputs
    if layer.name in layer_outputs:
        return layer_outputs[layer.name]

    if layer.name == starting_layer_name:
        out = layer(new_input)
        layer_outputs[layer.name] = out
        return out

    prev_layers = []
    for node in layer._inbound_nodes:
        if isinstance(node.inbound_layers, Iterable):
            prev_layers.extend(node.inbound_layers)
        else:
            prev_layers.append(node.inbound_layers)

    pl_outs = []
    for pl in prev_layers:
        pl_outs.extend([get_output_of_layer(
            pl, new_input, starting_layer_name)])

    out = layer(pl_outs[0] if len(pl_outs) == 1 else pl_outs)
    layer_outputs[layer.name] = out
    return out


# Returns a submodel for a specified input and output layer
def get_model(input_layer: int, output_layer: int):
    global selected_model

    layer_number = input_layer
    starting_layer_name = selected_model.layers[layer_number].name

    if input_layer == 0:
        new_input = selected_model.input

        return models.Model(new_input, selected_model.layers[output_layer].output)
    else:
        new_input = layers.Input(batch_shape=selected_model.get_layer(
            starting_layer_name).get_input_shape_at(0))

    new_output = get_output_of_layer(
        selected_model.layers[output_layer],
        new_input,
        starting_layer_name)
    model = models.Model(new_input, new_output)

    return model


# Navigates the model structure to find regions without parallel paths,
# returns valid split locations
def create_valid_splits():
    global selected_model
    model = selected_model

    layer_index = 1
    multi_output_count = 0

    valid_splits = []
    for layer in model.layers[1:]:

        if len(layer._outbound_nodes) > 1:
            multi_output_count += len(layer._outbound_nodes) - 1

        #if isinstance(layer._inbound_nodes[0].inbound_layers, list):
        if type(layer._inbound_nodes[0].inbound_layers) == list:
            if len(layer._inbound_nodes[0].inbound_layers) > 1:
                multi_output_count -= (
                    len(layer._inbound_nodes[0].inbound_layers) - 1)

        if multi_output_count == 0:
            valid_splits.append(layer_index)

        layer_index += 1

    return valid_splits


selected_model = None


def slice_model(model: models.Model) -> List[Slice]:
    global selected_model
    global layer_outputs
    layer_outputs = {}
    selected_model = model

    split_points = create_valid_splits()

    sliced_network: List[Slice] = []

    for index, split_point in enumerate(split_points):

        result = Slice()
        result.model = model

        if index == 0:
            first_point = 0
        else:
            first_point = split_points[index - 1] + 1

        result.input_layer = first_point
        result.output_layer = split_point

        result.model = get_model(first_point, split_point)
        sliced_network.append(result)

    return sliced_network


from docker.models.nodes import Node
from docker.models.images import Image
from docker.models.services import Service
from docker.types import RestartPolicy, ServiceMode, EndpointSpec, NetworkAttachmentConfig

from .api_common import PapdlAPIContext, get_papdl_service, get_service_status, get_service_node_ip_mapping
from .common import AppType
from ..slice.slice import Slice
from typing import List, TypedDict, Dict
from os import path, mkdir


class PapdlIperfAPI:
    def __init__(self, context: PapdlAPIContext):
        self.context = context
        self.service: Service = None

    def get_iperf_node_ip_mapping(self) -> Dict[str, str]:
        return get_service_node_ip_mapping(self.context, self.service)

    def spawn_iperfs(
        self,
    ) -> Service:
        iperf_services = get_papdl_service(
            self.context, labels={"type": AppType.IPERF.value})
        if (len(iperf_services) != 0):
            iperf_services[0].remove()

        self.context.logger.info(
            f"Spawning iperf3 network inspection service...")
        self.context.client.images.pull("networkstatic/iperf3")

        # es = EndpointSpec(ports={5201:5201})
        es = EndpointSpec(mode="vip", ports={5201: 5201})
        nac = NetworkAttachmentConfig(self.context.network.name)
        rp = RestartPolicy(condition="any")

        service = self.context.client.services.create(
            image="networkstatic/iperf3",
            name=f"iperf",
            mode=ServiceMode(mode="global"),
            endpoint_spec=es,
            networks=[nac],
            restart_policy=rp,
            args=["-s"],
            labels={
                "papdl": "true",
                "type": AppType.IPERF.value
            }
        )
        self.context.cleanup_target["services"].append(service)
        self.service = service
        return service



from shutil import copytree, copyfile
from typing import List, TypedDict, Generator, Dict, Callable, Union
from docker.models.images import Image
from docker.models.services import Service
from docker.models.secrets import Secret
from docker.models.networks import Network
from docker.models.nodes import Node
from .common import BenchmarkPreferences, AppType
from random_word import RandomWords
from getpass import getuser
from os import getuid, path
from tempfile import mkdtemp

import docker


class CleanupTarget(TypedDict):
    tempfolders: List[str]
    images: List[Image]
    services: List[Service]


class CertSpecs(TypedDict):
    C: str
    ST: str
    L: str
    O: str
    OU: str
    CN: str
    emailAddress: str
    validity_seconds: int


class PapdlAPIContext:
    def __init__(self, preference: BenchmarkPreferences = None, certSubject: CertSpecs = None):
        project_name: str
        self.client = docker.from_env()
        if preference is None:
            self.shell = True  # Shell class only to be used as a docker client to access utility methods
            return
        else:
            self.shell = False

        # r = RandomWords()
        # if project_name is None:
        #     # project_name = r.get_random_word()
        #     project_name = "debug"

        self.logger = preference["logger"]
        self.dircontext = {}
        self.local_user = getuser()
        self.local_uid = getuid()
        self.project_name = preference["project_name"]
        self.dircontext["api_module_path"] = path.dirname(
            path.abspath(__file__))
        self.cleanup_target = CleanupTarget(
            tempfolders=[], images=[], services=[])

        papdl_networks = get_papdl_network(self)
        if (len(papdl_networks) == 0):
            self.network:Network = self.client.networks.create(
                name=f"papdl_overlay",
                attachable=True,
                driver="overlay",
                labels={
                    "papdl": "true"
                }
            )
        else:
            self.network:Network = papdl_networks[0]

        self.preference = preference

        self.devices:List[Node] = list(
            filter(
                lambda n: n.attrs['Status']['State'] == 'ready',
                self.client.nodes.list()))
        self.logger.info(f"Available nodes: {self.devices}")

        # Dynamically set during runtime
        self.registry_service: Union[Service, None] = None
        self.iperf_service: Union[Service, None] = None
        self.benchmark_image: Union[Image, None] = None

        if (certSubject is None):
            self.cert_specs = CertSpecs(
                C="UK",
                ST="Fife",
                L="St. Andrews",
                O="University of St Andrews",
                OU="School of Computer Science",
                CN=getuser(),
                emailAddress=f"{getuser()}@st-andrews.ac.uk",
                validity_seconds=10 * 365 * 24 * 60 * 60
            )


def prepare_build_context(context: PapdlAPIContext) -> str:
    build_context = mkdtemp()
    context.cleanup_target["tempfolders"].append(build_context)
    return build_context


def copy_app(context: PapdlAPIContext, app_type: AppType,
             build_context: str) -> str:
    copytree(
        path.join(
            context.dircontext["api_module_path"],
            "containers",
            app_type.value,
            "app"),
        path.join(build_context, "app")
    )
    copyfile(
        path.join(
            context.dircontext["api_module_path"],
            "containers",
            app_type.value,
            "Dockerfile"),
        path.join(build_context, "Dockerfile")
    )
    copyfile(
        path.join(
            context.dircontext["api_module_path"],
            "containers",
            app_type.value,
            "requirements.txt"),
        path.join(build_context, "requirements.txt")
    )


def _get_papdl_component(list_handler: Callable,
                         labels: Dict[str, str], name=None) -> List:
    query = {}
    if name is not None:
        query["name"] = name
    query["label"] = ["papdl=true"]
    if (len(labels.keys()) != 0):
        for k, v in labels.items():
            query["label"].append(f"{k}={v}")
    return list_handler(filters=query)


def get_papdl_secret(context: PapdlAPIContext,
                     labels: Dict[str, str] = {}, name=None) -> List[Secret]:
    return _get_papdl_component(context.client.secrets.list, labels, name)


def get_papdl_service(context: PapdlAPIContext,
                      labels: Dict[str, str] = {}, name=None):
    return _get_papdl_component(context.client.services.list, labels, name)


def get_papdl_image(context: PapdlAPIContext,
                    labels: Dict[str, str] = {}, name=None):
    return _get_papdl_component(context.client.images.list, labels, name)


def get_papdl_network(context: PapdlAPIContext,
                      labels: Dict[str, str] = {}, name=None):
    return _get_papdl_component(context.client.networks.list, labels, name)


def get_service_status(service: Service) -> List[str]:
    return list(map(lambda s: s['Status']['State'], service.tasks()))


def get_service_node_ip_mapping(context: PapdlAPIContext, service: Service):
    service_tasks = service.tasks()
    mapping = {}
    for task in service_tasks:
        node_id = task['NodeID']
        task_networks = task["NetworksAttachments"]
        project_network_config = list(
            filter(
                lambda n: n['Network']['Spec']['Name'] == context.network.name,
                task_networks))[0]
        task_ip = project_network_config['Addresses'][0]
        ip = task_ip.split("/")[0]
        mapping[node_id] = ip
    return mapping


def get_docker_logs(context: PapdlAPIContext,
                    log_generator: Generator[Dict, None, None]):
    for chunk in log_generator:
        if 'stream' in chunk:
            for line in chunk["stream"].splitlines():
                context.logger.debug(line)




import heapq
from json import loads, dumps
from enum import Enum
from typing import Dict, List, TypedDict, Union, Set, Tuple, NamedTuple
from copy import deepcopy
import pandas as pd
from collections import deque
from functools import reduce

test_input = """
{
   "0jv7aoa9p5ll1zxqca7852pkp":{
      "model_performance":{
         "model_5":{
            "benchmark_time":0.0009600341320037842,
            "benchmark_size":1328
         },
         "model_9":{
            "benchmark_time":0.0007116296291351319,
            "benchmark_size":168
         },
         "model_0":{
            "benchmark_time":0.0012714905738830566,
            "benchmark_size":928
         },
         "model_6":{
            "benchmark_time":0.000756922960281372,
            "benchmark_size":1328
         },
         "model_4":{
            "benchmark_time":0.0009644310474395752,
            "benchmark_size":1728
         },
         "model_3":{
            "benchmark_time":0.0009650299549102783,
            "benchmark_size":2128
         },
         "model_1":{
            "benchmark_time":0.0007647640705108642,
            "benchmark_size":1328
         },
         "model_8":{
            "benchmark_time":0.0007638306617736816,
            "benchmark_size":528
         },
         "model_2":{
            "benchmark_time":0.0009319167137145997,
            "benchmark_size":1728
         },
         "model_7":{
            "benchmark_time":0.0007658138275146484,
            "benchmark_size":928
         }
      },
      "network_performance":{
         "0jv7aoa9p5ll1zxqca7852pkp":{
            "bandwidth":{
               "sent_bps":27967170054.26285,
               "received_bps":26865340762.208767
            },
            "latency":{
               "rtt_min_ms":0.0,
               "rtt_avg_ms":0.0,
               "rtt_max_ms":0.02
            }
         },
         "u4epygnre7zj60jmtz6nfoqh3":{
            "bandwidth":{
               "sent_bps":916772831.8971637,
               "received_bps":856761316.6488367
            },
            "latency":{
               "rtt_min_ms":0.24,
               "rtt_avg_ms":0.35,
               "rtt_max_ms":0.54
            }
         },
         "g70ppy9wwa09w5x5al3rqhkk8":{
            "bandwidth":{
               "sent_bps":930054462.7564534,
               "received_bps":868669113.067231
            },
            "latency":{
               "rtt_min_ms":0.27,
               "rtt_avg_ms":0.47,
               "rtt_max_ms":1.5
            }
         }
      }
   },
   "g70ppy9wwa09w5x5al3rqhkk8":{
      "model_performance":{
         "model_6":{
            "benchmark_time":0.0008424403667449951,
            "benchmark_size":1328
         },
         "model_9":{
            "benchmark_time":0.0009128034114837646,
            "benchmark_size":168
         },
         "model_5":{
            "benchmark_time":0.0014638190269470215,
            "benchmark_size":1328
         },
         "model_3":{
            "benchmark_time":0.00109647536277771,
            "benchmark_size":2128
         },
         "model_1":{
            "benchmark_time":0.0008775646686553955,
            "benchmark_size":1328
         },
         "model_7":{
            "benchmark_time":0.0008452577590942383,
            "benchmark_size":928
         },
         "model_4":{
            "benchmark_time":0.0008671987056732178,
            "benchmark_size":1728
         },
         "model_8":{
            "benchmark_time":0.0008331222534179687,
            "benchmark_size":528
         },
         "model_0":{
            "benchmark_time":0.0012930092811584473,
            "benchmark_size":928
         },
         "model_2":{
            "benchmark_time":0.0008685297966003418,
            "benchmark_size":1728
         }
      },
      "network_performance":{
         "0jv7aoa9p5ll1zxqca7852pkp":{
            "bandwidth":{
               "sent_bps":882522.8637524123,
               "received_bps":53595.046866913304
            },
            "latency":{
               "rtt_min_ms":0.28,
               "rtt_avg_ms":0.62,
               "rtt_max_ms":1.92
            }
         },
         "u4epygnre7zj60jmtz6nfoqh3":{
            "bandwidth":{
               "sent_bps":906944414.0616016,
               "received_bps":851076057.6263435
            },
            "latency":{
               "rtt_min_ms":0.31,
               "rtt_avg_ms":0.41,
               "rtt_max_ms":1.43
            }
         },
         "g70ppy9wwa09w5x5al3rqhkk8":{
            "bandwidth":{
               "sent_bps":11752377110.892979,
               "received_bps":11277505010.402567
            },
            "latency":{
               "rtt_min_ms":0.0,
               "rtt_avg_ms":0.0,
               "rtt_max_ms":0.02
            }
         }
      }
   },
   "u4epygnre7zj60jmtz6nfoqh3":{
      "model_performance":{
         "model_8":{
            "benchmark_time":0.0005680563449859619,
            "benchmark_size":528
         },
         "model_1":{
            "benchmark_time":0.0005618522167205811,
            "benchmark_size":1328
         },
         "model_4":{
            "benchmark_time":0.0005836150646209717,
            "benchmark_size":1728
         },
         "model_0":{
            "benchmark_time":0.0008509521484375,
            "benchmark_size":928
         },
         "model_3":{
            "benchmark_time":0.0005839769840240478,
            "benchmark_size":2128
         },
         "model_7":{
            "benchmark_time":0.0005602047443389892,
            "benchmark_size":928
         },
         "model_2":{
            "benchmark_time":0.0005665426254272461,
            "benchmark_size":1728
         },
         "model_9":{
            "benchmark_time":0.0005675253868103027,
            "benchmark_size":168
         },
         "model_5":{
            "benchmark_time":0.0005712642669677734,
            "benchmark_size":1328
         },
         "model_6":{
            "benchmark_time":0.0005655736923217774,
            "benchmark_size":1328
         }
      },
      "network_performance":{
         "0jv7aoa9p5ll1zxqca7852pkp":{
            "bandwidth":{
               "sent_bps":882469.9762686674,
               "received_bps":55801.47766144708
            },
            "latency":{
               "rtt_min_ms":0.19,
               "rtt_avg_ms":0.29,
               "rtt_max_ms":0.67
            }
         },
         "u4epygnre7zj60jmtz6nfoqh3":{
            "bandwidth":{
               "sent_bps":33234176155.877346,
               "received_bps":33233245784.088963
            },
            "latency":{
               "rtt_min_ms":0.0,
               "rtt_avg_ms":0.0,
               "rtt_max_ms":0.02
            }
         },
         "g70ppy9wwa09w5x5al3rqhkk8":{
            "bandwidth":{
               "sent_bps":917023975.5235491,
               "received_bps":891952506.9651792
            },
            "latency":{
               "rtt_min_ms":0.23,
               "rtt_avg_ms":0.43,
               "rtt_max_ms":1.54
            }
         }
      }
   }
}"""

test = loads(test_input)


class Model():
    def __init__(self, name):
        self.name = name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, Model):
            return False
        return other.name == self.name

    def __str__(self):
        return self.name


class Device():
    def __init__(self, name):
        self.name = name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, Device):
            return other.name == self.name
        else:
            return False

    def __str__(self):
        return self.name


devices: List[Device] = [Device(k) for k in list(test.keys())]
source_device = devices[0]

models: List[Model] = [Model(m) for m in sorted(
    list(test[source_device.name]['model_performance'].keys()))]

input_size = 100


class Node():
    def __init__(self, model: Model = None, device: Device = None,
                 paths: List["Path"] = None):
        self.model: Model = model
        self.device: Device = device
        self.paths: List["Path"] = paths

    def __hash__(self) -> int:
        if self.model is None:
            return hash("NULL" + "-" + self.device.name)
        else:
            return hash(self.model.name + "-" + self.device.name)

    def __eq__(self, other: "Node") -> bool:
        if isinstance(other, Node):
            if self.model is None and other.model is not None:
                return False

            if other.model is None and self.model is not None:
                return False

            if self.model is None and other.model is None:
                return self.device.name == other.device.name

            return self.model.name == other.model.name and self.device.name == other.device.name
        return False

    def __str_children(self) -> str:
        result = []
        p: "Path"
        for p in self.paths:
            model_name = p['node'].model.name if p['node'].model is not None else "NULL"
            result.append(
                f"*{model_name}-{p['node'].device.name}-{p['penalty_ms']}*")
        return "\n".join(result)

    def __detailstr__(self) -> str:
        model_name = "NULL" if self.model is None else self.model.name
        return f"<NODE model:{model_name} device:{self.device.name} children=[\n{self.__str_children()}]>"

    def __str__(self) -> str:
        model_name = "NULL" if self.model is None else self.model.name
        return f"<NODE model:{model_name} device:{self.device.name}]>"


class Path(TypedDict):
    node: Node
    penalty_ms: float


visit_node_map: Dict[Node, Node] = {}


def calculate_performance_penalty(destination: Device, model: Model):
    return test[destination.name]["model_performance"][model.name]["benchmark_time"]


def calculate_network_penalty(
        source: Device, destination: Device, filesize_to_send: int):
    stats = test[source.name]["network_performance"][destination.name]
    latency = stats["latency"]["rtt_avg_ms"] / 1000  # get in seconds
    bandwidth = stats["bandwidth"]["sent_bps"]
    return (latency + (filesize_to_send / bandwidth)) * 1000


def calculate_network_penalty_from_model(
        source: Device, destination: Device, model: Model):
    filesize_to_send = test[source.name]["model_performance"][model.name]["benchmark_size"]
    return calculate_network_penalty(source, destination, filesize_to_send)


def generate_path(source: Device, destination: Device,
                  next_model: Model) -> Path:

    penalty: float
    if next_model is not None:
        # Generating for first node
        penalty = calculate_network_penalty_from_model(
            source, destination, next_model)
        penalty += calculate_performance_penalty(destination, next_model)
    else:
        penalty = calculate_network_penalty(source, destination, input_size)

    # If node was previously computed
    temp: Node = Node(model=next_model, device=destination)
    path: Path
    if temp in visit_node_map:
        path = Path(node=visit_node_map.get(temp), penalty_ms=penalty)
    else:
        path = Path(
            node=Node(
                model=next_model,
                device=destination,
                paths=[]),
            penalty_ms=penalty)
    return path


def rec_construct_path(models_left: List[Model], currNode: Node):
    # Check for duplicate node construct here
    if currNode in visit_node_map:
        return

    visit_node_map[currNode] = currNode
    if len(models_left) == 0:
        currNode.paths = [
            generate_path(
                source=currNode.device,
                destination=source_device,
                next_model=None)
        ]
        return
    else:
        paths = [
            generate_path(
                currNode.device,
                d,
                models_left[0]) for d in devices]
        currNode.paths = paths
        path: Path
        for path in currNode.paths:
            nextNode = path["node"]
            rec_construct_path(models_left=models_left[1:], currNode=nextNode)


head = Node(
    model=None,
    device=source_device,
)
# visit_node_map[head] = head
rec_construct_path(models_left=models, currNode=head)

# for n in sorted(visit_node_map.values(),key=lambda n: "n" if n.model is None else n.model.name ):
# print(n)
# print("")


class SearchStatus(NamedTuple):
    total_distance: float
    path: Path


class OptimalPath(NamedTuple):
    path: List[Node]
    penalty: float


class SearchCriterion(TypedDict):
    layer_must_be_in_device: Dict[Model, Device]
    layer_must_not_be_in_device: Dict[Model, Device]


def valid_path(path: List[Node], criteria: SearchCriterion):
    for node in path:
        for model, device in criteria["layer_must_be_in_device"].items():
            if model == node.model and device != node.device:
                return False

    for node in path:
        for model, device in criteria["layer_must_not_be_in_device"].items():
            if model == node.model and device == node.device:
                return False

    return True


def find_shortest_loop(
        start_node: Node, search_criteria: SearchCriterion) -> Union[OptimalPath, None]:
    visited = {start_node: [start_node]}
    queue: List[SearchStatus] = [
        SearchStatus(
            0,
            Path(
                node=start_node,
                penalty_ms=0))]
    while queue:
        total_penalty_ms, path = heapq.heappop(queue)
        current_node = path["node"]
        for child_path in current_node.paths:
            child_node = child_path["node"]
            if child_node not in visited:
                new_path = visited[current_node] + [child_node]  # Make a copy
                if valid_path(new_path, criteria=search_criteria):
                    visited[child_node] = new_path
                    new_penalty = total_penalty_ms + child_path["penalty_ms"]
                    heapq.heappush(
                        queue,
                        SearchStatus(
                            total_distance=new_penalty, path=Path(
                                node=child_node, penalty_ms=new_penalty))
                    )
            # Found a loop
            elif child_node == start_node and len(visited[current_node]) > 1:
                return OptimalPath(
                    path=visited[current_node] + [start_node], penalty=child_path["penalty_ms"])
    return None  # No loop found


sc = SearchCriterion(
    layer_must_be_in_device={
        models[-2]: source_device, models[-1]: devices[2]},
    layer_must_not_be_in_device={
        models[0]: source_device,
        models[1]: devices[1],
        models[2]: devices[2]}
)

result = find_shortest_loop(start_node=head, search_criteria=sc)

print("Devices: \n" + ", ".join([d.name for d in devices]))
print("Models: \n" + ", ".join([m.name for m in models]))
print("Constraints:")
for m, d in sc["layer_must_be_in_device"].items():
    print(f"Model {m.name} must be in device {d.name}")
for m, d in sc["layer_must_not_be_in_device"].items():
    print(f"Model {m.name} must not be in device {d.name}")

if result is None:
    print("No path found")
    exit(0)


class Slice(NamedTuple):
    models: List[Model]
    slice_index: Tuple[int, int]
    device: Device


def generate_slices(op: OptimalPath) -> List[Slice]:
    l = 1
    r = 2

    slices: List[Slice] = []
    while r < len(op.path):
        if op.path[r].device != op.path[l].device:
            nodes_slice = op.path[l:r]
            s = Slice(
                models=[
                    n.model for n in nodes_slice],
                slice_index=(
                    l,
                    r),
                device=op.path[l].device)
            slices.append(s)
            l = r
        r += 1
    return slices


print("Optimal Path: \n" +
      ",\n".join([f"<MODEL: {n.model} DEVICE: {n.device}>" for n in result.path]))
print("Penalty: " + str(result.penalty))

slices = generate_slices(result)

print("OPTIMAL SLICES: ")
for s in slices:
    print(
        f"<SLICE model: {', '.join( [m.name for m in s.models])}, device: {s.device}>")



from enum import Enum
from typing import TypedDict, Dict, List, Union, NamedTuple, Tuple
import tqdm
import logging
from colorama import Fore, Style
from time import time, sleep
from threading import Thread
from docker.models.services import Service
from docker.models.images import Image
from docker.models.networks import Network
from docker.models.containers import Container
from docker.models.nodes import Node
import coloredlogs




class ContainerBehaviourException(Exception):
    def __init__(self, message: str, service: Service = None):
        self.message = message
        self.service = service
        super().__init__(self.message)


class AppType(Enum):
    ORCHESTRATOR = "Orchestrator"
    BENCHMARKER = "Benchmarker"
    SLICE = "Slice"
    IPERF = "Iperf"
    REGISTRY = "Registry"


class SplitStrategy(Enum):
    ATOMIC = "atomic"
    SCISSION = "scission"
    SCISSION_TL = "scission_tl"

    @staticmethod
    def from_str(label):
        if label == 'scission':
            return SplitStrategy.SCISSION
        elif label == 'scission_tl':
            return SplitStrategy.SCISSION_TL
        elif label == 'atomic':
            return SplitStrategy.ATOMIC
        else:
            raise NotImplementedError


class TqdmLoggingHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except Exception:
            self.handleError(record)


class ColourFormatter(logging.Formatter):
    format = "[%(asctime)s - %(levelname)s ] %(message)s (%(filename)s:%(lineno)d)"
    FORMATS = {
        logging.DEBUG: Fore.BLUE + format + Fore.RESET,
        logging.INFO: Fore.YELLOW + format + Fore.RESET,
        logging.ERROR: Fore.RED + format + Fore.RESET
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def prepare_logger(level=logging.DEBUG) -> logging.Logger:
    # logger = logging.getLogger(__name__)
    # logger.setLevel(level)

    # ch = logging.StreamHandler()
    # ch.setLevel(level=level)
    # ch.setFormatter(ColourFormatter())
    # logger.addHandler(ch)
    # logger.propagate = False
    logger = logging.getLogger(__name__)
    logger.propagate = False
    coloredlogs.install(level=level,logger=logger)
    return logger


class PapdlException(Exception):
    def __init__(self,message):
        self.message = message

class BenchmarkPreferences(TypedDict):
    project_name:str

    service_idle_detection: int
    startup_timeout: int
    split_strategy: SplitStrategy
    logger: logging.Logger

    model_test_batch_size:int
    model_test_number_of_repeats:int
    bandwidth_test_duration_sec:int
    latency_test_count:int
    free_memory_multiplier:float



##########


class BenchmarkSetup(TypedDict):
    project_name: str
    registry_service: Service
    iperf_service: Service
    network: Network
    benchmark_image: Image


class NodeBenchmarkMetadata(TypedDict):
    node: Node
    task: Dict
    raw_log: str
    host_ip: str
    docker_dns_ip: str


class NodeBenchmark(TypedDict):
    result: Dict
    metadata: NodeBenchmarkMetadata


class PapdlTest(TypedDict):
    benchmark: List[NodeBenchmark]
    benchmark_setup: BenchmarkSetup



import logging
import os
import keras as k


class ServerConfig:
    def __init__(self):
        self.server_logger = logging.getLogger("server")
        server_logger_handler: logging.StreamHandler = logging.StreamHandler()
        server_logger_handler.setFormatter(
            logging.Formatter("%(asctime)s [%(name)-6s] %(levelname)-8s %(message)s"
                              ))
        self.server_logger.addHandler(server_logger_handler)
        self.server_logger.setLevel(logging.DEBUG)

        self.FORWARD_URL = os.getenv("FORWARD_URL")
        self.FOREWARD_PORT = 8765



import click
from .slice import main as slicer
from .configure import main as configurer
from .clean import main as cleaner
from .benchmark import main as benchmarker
from .deploy import main as deployer
import os


os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

@click.group()
def main():
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    pass


main.add_command(slicer.slice)
main.add_command(configurer.configure)
main.add_command(cleaner.clean)
main.add_command(benchmarker.benchmark)
main.add_command(deployer.deploy)



from typing import Dict, List, Callable, Optional, NamedTuple
from ._api_benchmark import PapdlBenchmarkAPI
from ._api_iperf import PapdlIperfAPI
from ._api_registry import PapdlRegistryAPI
from .api_common import PapdlAPIContext, get_papdl_service, get_service_status
from docker.models.services import Service
from docker.models.nodes import Node
from .common import ContainerBehaviourException
from ..slice.slice import Slice
from time import time, sleep
from ..backend.common import BenchmarkSetup, NodeBenchmarkMetadata, NodeBenchmark, PapdlTest


import json


class DeploymentStatus(NamedTuple):
    node_service_mapping: Dict[Node, Service]
    node_ip_mapping: Dict[str, str]


class PapdlAPI:
    def __init__(self, context: PapdlAPIContext):
        self.context = context
        self.benchmark_api = PapdlBenchmarkAPI(context)
        self.registry_api = PapdlRegistryAPI(context)
        self.iperf_api = PapdlIperfAPI(context)

    def cleanup(self):
        pass

    def _all_match_any_of(_all: List[str], _any: List[str]) -> bool:
        return all([True if x in _any else False for x in _all])

    # make timeout on into an enum, turn into sets with all possible states
    def _deploy_service_with_timeout(
            self, service_spawner: Callable, timeout_on: List[str], service_spawner_args={}) -> Service:
        service = service_spawner(**service_spawner_args)
        self.context.logger.info(f"Spawing service {service.name} ...")
        start_time = time()
        while True:
            if time() - \
                    start_time > self.context.preference['startup_timeout']:
                raise ContainerBehaviourException(
                    f"Service {service.name} timed out trying to spawn...", service=service)
            else:
                service_status = get_service_status(service)
                self.context.logger.debug(
                    f"Service {service.name} has status: {service_status}")
                if PapdlAPI._all_match_any_of(service_status, timeout_on):
                    break
                if PapdlAPI._all_match_any_of(
                        service_status, ["failed", "shutdown", "rejected", "orphaned", "remove"]):
                    raise ContainerBehaviourException(
                        f"Service {service.name} was/has {service_status}", service=service)
                sleep(1)
        return service

    def deploy_benchmarkers(self, slices: List[Slice]) -> DeploymentStatus:

        registry_service = self._deploy_service_with_timeout(
            self.registry_api.spawn_registry, ["running", "complete"])
        iperf_service = self._deploy_service_with_timeout(
            self.iperf_api.spawn_iperfs, ["running", "complete"])
        self.context.registry_service = registry_service
        self.context.iperf_service = iperf_service
        iperf_node_ip_mapping = self.iperf_api.get_iperf_node_ip_mapping()

        self.context.logger.info(
            f"Node to docker-dns-ip mapping: {iperf_node_ip_mapping}")

        image = self.benchmark_api.build_benchmark_image(slices)
        node: Node
        deployed_services: Dict[Node, Service] = {}
        self.context.logger.info(f"Running benchmarking on nodes: {self.context.devices}")
        for node in self.context.devices:
            deployed_services[node] = self._deploy_service_with_timeout(
                self.benchmark_api.spawn_benchmarker_on_node,
                ["complete"],
                {'image': image, 'node': node,
                    'iperf_test_ips': iperf_node_ip_mapping.values()}
            )
        return DeploymentStatus(
            node_service_mapping=deployed_services,
            node_ip_mapping=iperf_node_ip_mapping)

    def get_raw_service_logs(self, service: Service) -> str:
        log_read_start = time()
        lines = []
        while True:
            state = service.tasks()[0]['Status']['State']
            if state == 'complete':
                log_line: bytes
                for log_line in service.logs(stdout=True):
                    log_decoded = log_line.decode('utf-8')
                    lines.append(log_decoded)
            else:
                if time() - \
                        log_read_start > self.context.preference["service_idle_detection"]:
                    raise ContainerBehaviourException(
                        f"Log reading timed out on service {service.name}", service=Service)

    def get_service_logs(self, service: Service) -> Dict:
        log_read_start = time()
        while True:
            state = service.tasks()[0]['Status']['State']
            if state == 'complete':
                log_line: bytes
                self.context.logger.debug(
                    f"Retrieved the following logs from the service {service.name}")
                for log_line in service.logs(stdout=True):
                    log_decoded = log_line.decode('utf-8')
                    self.context.logger.debug(log_decoded)
                    if log_decoded[:11] == '[BENCHMARK]':
                        message = log_decoded[11:]
                        return json.loads(message)
            else:
                if time() - \
                        log_read_start > self.context.preference['service_idle_detection']:
                    raise ContainerBehaviourException(
                        f"Log reading timed out on service {service.name}", service=service)



import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import click
from .slice import slice_model
import traceback
from ..backend.common import prepare_logger
from logging import DEBUG
import keras
import dill as pickle


@click.command()
@click.argument("model_path")
@click.option("-o", "--output", default=None)
def slice(
        model_path: str,
        output:str = "slice.json"
):

    model = None
    logger = prepare_logger(DEBUG)
    try:
        model = keras.models.load_model(model_path)
    except Exception:
        logger.error(traceback.format_exc())
        exit(1)

    logger.info("Loading Model...")
    logger.info(model)
    logger.info("Slicing Model...")

    sliced_model = slice_model(model)

    if(output is None):
        output="slices.pickle"

    try:
        with open(output, "wb") as f:
            pickle.dump(sliced_model,f,protocol=pickle.HIGHEST_PROTOCOL)
        logger.info(f"Saving sliced model list as '{output}'")
    except Exception as e:
        logger.error(traceback.format_exc())
        exit(1)



import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import click
from ..backend.common import prepare_logger
import logging
from ..configure.configure import Configuration,Configurer
import traceback
from .deploy import deploy_configuration
import dill as pickle


@click.command()
@click.argument("configuration_path")
@click.option("-d","--debug",type=bool,default=False)
def deploy(
    configuration_path:str,
    debug:bool
):
    try:
        logger = prepare_logger(logging.DEBUG)
        logging.info("Deploying models...")
        configuration:Configuration = None
        with open(configuration_path,"rb") as f:
            configuration = pickle.load(f)
        logger.info(configuration["blocks"])
        deploy_configuration(configuration,debug)
    except Exception:
        logger.error(traceback.format_exc())
        exit(1)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: iss_message.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import ndarray_pb2 as ndarray__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='iss_message.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x11iss_message.proto\x1a\rndarray.proto\"X\n\nIssMessage\x12\x16\n\trequestId\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1b\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x08.ndarrayH\x01\x88\x01\x01\x42\x0c\n\n_requestIdB\x07\n\x05_datab\x06proto3'
  ,
  dependencies=[ndarray__pb2.DESCRIPTOR,])




_ISSMESSAGE = _descriptor.Descriptor(
  name='IssMessage',
  full_name='IssMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='requestId', full_name='IssMessage.requestId', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='IssMessage.data', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_requestId', full_name='IssMessage._requestId',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_data', full_name='IssMessage._data',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=36,
  serialized_end=124,
)

_ISSMESSAGE.fields_by_name['data'].message_type = ndarray__pb2._NDARRAY
_ISSMESSAGE.oneofs_by_name['_requestId'].fields.append(
  _ISSMESSAGE.fields_by_name['requestId'])
_ISSMESSAGE.fields_by_name['requestId'].containing_oneof = _ISSMESSAGE.oneofs_by_name['_requestId']
_ISSMESSAGE.oneofs_by_name['_data'].fields.append(
  _ISSMESSAGE.fields_by_name['data'])
_ISSMESSAGE.fields_by_name['data'].containing_oneof = _ISSMESSAGE.oneofs_by_name['_data']
DESCRIPTOR.message_types_by_name['IssMessage'] = _ISSMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

IssMessage = _reflection.GeneratedProtocolMessageType('IssMessage', (_message.Message,), {
  'DESCRIPTOR' : _ISSMESSAGE,
  '__module__' : 'iss_message_pb2'
  # @@protoc_insertion_point(class_scope:IssMessage)
  })
_sym_db.RegisterMessage(IssMessage)


# @@protoc_insertion_point(module_scope)



from docker.models.nodes import Node
from docker.models.images import Image
from docker.models.services import Service
from docker.types import RestartPolicy, NetworkAttachmentConfig

from .api_common import PapdlAPIContext, prepare_build_context, copy_app, get_docker_logs, get_service_node_ip_mapping
from .common import AppType
from ..slice.slice import Slice
from typing import List, TypedDict
from os import path, mkdir


class PapdlBenchmarkAPI:
    def __init__(self, context: PapdlAPIContext):
        self.context = context
        self.services = []

    def prepare_benchmark_build(self, slices: List[Slice]) -> str:
        self.context.logger.info(f"Preparing benchmark build...")
        build_context = prepare_build_context(self.context)
        model_paths = path.join(build_context, "models")
        mkdir(model_paths)

        slice: Slice
        for slice in slices:
            model = slice.model
            model_path = path.join(model_paths, slice.model.name)
            mkdir(model_path)
            model.save(model_path, overwrite=True)

        copy_app(self.context, AppType.BENCHMARKER, build_context)
        return build_context

    def get_benchmark_node_ip_mapping(self):
        result = {}
        s: Service
        for s in self.services:
            s_node_ip_mapping = get_service_node_ip_mapping(self.context, s)
            result.update(s_node_ip_mapping)

        return result

    def build_benchmark_image(self, slices: List[Slice]) -> Image:

        name = f"localhost:443/{self.context.project_name}/benchmark"
        buildargs = {
            "local_user": self.context.local_user,
            "local_uid": str(self.context.local_uid),
        }
        build_context = self.prepare_benchmark_build(slices)

        self.context.logger.info(
            f"Building benchmark with image {name} and context {build_context}")
        image, build_logs = self.context.client.images.build(
            path=build_context,
            tag=name,
            buildargs=buildargs,
            labels={
                "papdl": "true",
                "project_name": self.context.project_name,
                "type": AppType.BENCHMARKER.value
            }
        )
        get_docker_logs(self.context, build_logs)
        self.context.cleanup_target["images"].append(image)

        self.context.logger.info(f"Pushing image {name} to registry for distribution...")
        self.context.client.images.push(name)
        self.context.benchmark_image = image
        return image

    def spawn_benchmarker_on_node(
        self,
        image: Image,
        node: Node,
        iperf_test_ips: List[str]
    ) -> Service:
        image_name = image.tags[0]
        self.context.logger.info(f"Spawning service for image {image_name}")
        # node_ips = list(map(lambda n: n.attrs["Status"]["Addr"],self.context.devices))
        nac = NetworkAttachmentConfig(self.context.network.name)
        
        benchmark_prefs = self.context.preference
        self.context.logger.info(benchmark_prefs)

        service = self.context.client.services.create(
            image=image_name,
            command=f"python3 -m server",
            name=f"{self.context.project_name}_benchmark_{node.id}",
            constraints=[f"node.id=={node.id}"],
            # user=f"{self.context.local_user}:{self.context.local_user",
            # user=self.context.local_user,
            restart_policy=RestartPolicy(condition="none"),
            env=[
                f"PAPDL_WORKERS={' '.join(iperf_test_ips)}",
                f"MODEL_TEST_BATCH_SIZE={benchmark_prefs['model_test_batch_size']}",
                f"MODEL_TEST_NUMBER_OF_REPEATS={benchmark_prefs['model_test_number_of_repeats']}",
                f"BANDWIDTH_TEST_DURATION_SEC={benchmark_prefs['bandwidth_test_duration_sec']}",
                f"LATENCY_TEST_COUNT={benchmark_prefs['latency_test_count']}",
                f"FREE_MEMORY_MULTIPLIER={benchmark_prefs['free_memory_multiplier']}"
            ],
            labels={
                "papdl": "true",
                "project_name": self.context.project_name,
                "type": AppType.BENCHMARKER.value
            },
            networks=[nac]
        )
        self.context.cleanup_target["services"].append(service)
        self.services.append(service)
        return service




import os
import json
import sys

from websockets.client import connect as ws_connect
from websockets.server import serve as ws_serve
from websockets.datastructures import Headers,HeadersLike
from http import HTTPStatus
import numpy as np
import io
from typing import Optional,Dict,Tuple,Union,List
import logging
import asyncio
import traceback
from asyncio import Future
os.environ["TF_CPP_MIN_LOG_LEVEL"] = '3'
import uuid
import requests
import aiohttp
import uproot
from time import time_ns
import asyncstdlib as asynclib
import lz4.frame
from sanic.log import logger

from websockets.legacy.client import Connect

from sanic import Request,Websocket,Sanic
from sanic import response


def prepare_logger()->logging.Logger:
    logger = logging.getLogger("server")
    server_logger_handler: logging.StreamHandler = logging.StreamHandler()
    server_logger_handler.setFormatter(
        logging.Formatter("%(asctime)s [%(name)-6s] %(levelname)-8s %(message)s"
                            ))
    logger.addHandler(server_logger_handler)
    logger.setLevel(logging.DEBUG)
    return logger

# forward_connection:Union[Connect,None] = None
# forward_url:Union[str,None] = None

CURR_HOST = "0.0.0.0"
CURR_HOST_PORT = 8765
print("UPROOT SERIALIZATION")

def get_next_url()->str:
    forward_service_name = os.environ.get("FORWARD")
    logger.info(f"Loaded forward url: {forward_service_name}")
    return f"ws://{forward_service_name}:8765"

def get_all_slices()->List[str]:
    service_list = os.environ.get("SLICES").split(",")
    logger.info(f"Loaded slice services: {service_list}")
    return service_list

def get_input_dims()->Tuple[int]:
    input_dims_list = (int(s) for s in os.environ.get("INPUTDIMS").split(","))
    return tuple(input_dims_list)
    

app = Sanic("OrchestratorServer")
app.ctx.forward_url:str = get_next_url()
app.ctx.all_slice_services:List[str] = get_all_slices()
app.ctx.forward_connection:Union[Connect,None] = None
app.ctx.rrm_lock = asyncio.Lock()
app.ctx.request_response_map:Dict[str,Future] = {}
app.ctx.input_dims = get_input_dims()
app.config.WEBSOCKET_MAX_SIZE = sys.maxsize
app.config.KEEP_ALIVE_TIMEOUT = sys.maxsize
app.config.RESPONSE_TIMEOUT = sys.maxsize
app.config.RESPONSE_TIMEOUT = sys.maxsize

# request_response_map:Dict[str,Future] = {}
# rrm_lock = asyncio.Lock()
@app.websocket("/predict")
async def record_response(request:Request,ws:Websocket):
    async for data in ws:
        logger.info("RECEIVED PREDICT!")
        requestId:str
        try:
            decompressed_buff = io.BytesIO()
            decompressed_buff.write(lz4.frame.decompress(data))
            decompressed_buff.seek(0)
            requestId = decompressed_buff.read(36).decode("utf-8")
            logger.info("DECOMPRESSED!")
            logger.info(requestId)
            # data = np.load(decompressed_buff)
            # logger.info("LOADED!")
            # logger.info(data.shape)
            async with app.ctx.rrm_lock:
                app.ctx.request_response_map[requestId].set_result(decompressed_buff.read())
        except Exception as e:
            logger.error(traceback.format_exc())
            async with app.ctx.rrm_lock:
                app.ctx.request_response_map[requestId].set_exception(e)

@app.post("/input")
async def make_prediction(request:Request):
    logger.info("INPUT REQUEST!")
    try:
        request_buff = io.BytesIO()
        request_buff.write(request.body)
        request_buff.seek(0)
        array:np.ndarray = np.load(request_buff)
        expected_input_dims = app.ctx.input_dims

        if len(expected_input_dims) + 1 != len(array.shape) or array.shape[1:] != expected_input_dims:
            raise ValueError()
        
        send_buff = io.BytesIO()
        request_id = str(uuid.uuid4())
        send_buff.write(bytes(request_id,"utf-8"))
        send_buff.write(request.body)
        send_buff.seek(0)
        
        
        loop = asyncio.get_running_loop()
        response_future = loop.create_future()
        async with app.ctx.rrm_lock:
            app.ctx.request_response_map[request_id] = response_future
        await app.ctx.forward_connection.send(lz4.frame.compress(send_buff.read()))
        logger.info("WAITING!")
        await response_future
        return response.raw(body=response_future.result(), status = HTTPStatus.OK)
    except ValueError as e:
        logger.error(traceback.format_exc())
        return response.raw(body="Input dimensions does not match the model",status=HTTPStatus.BAD_REQUEST)
    except Exception as e:
        logger.error(traceback.format_exc())
        return response.json(body=e,status=HTTPStatus.INTERNAL_SERVER_ERROR)

    
@app.get("/connect")
async def connect_to_forward(request:Request):
    try:
        app.ctx.forward_connection:Connect = await ws_connect(f"{app.ctx.forward_url}/predict", max_size=sys.maxsize)
        logger.info(f"Successfly connected to forward_url: {app.ctx.forward_url}")
        return response.text("Successfully connected to  forward url",status=HTTPStatus.OK)
    except:
        logger.error(traceback.format_exc())
        return response.text("Unable to connect to forward url")

@app.get("/activate")
async def activate_all_slices(request:Request):
    try:
        result = {}
        service_name:str
        for service_name in app.ctx.all_slice_services:
            slice_forward_connect_url = f"http://{service_name}:8765/connect"
            async with aiohttp.ClientSession() as session:
                async with session.get(slice_forward_connect_url) as resp:
                    result[service_name] = {
                        "status" : resp.status,
                        "response": await resp.text()
                    }
        return response.json(result,status=HTTPStatus.OK) 
    except Exception:
        logger.error(traceback.format_exc())
        return response.text(body="Failed to activate one or all slice nodes",status=HTTPStatus.BAD_REQUEST)


@app.get("/healthcheck")
async def perform_healthcheck(request:Request):
    logger.info(f"Generating healthcheck")
    return response.json({"forward_url":app.ctx.forward_url, "connected": app.ctx.forward_connection is not None and app.ctx.forward_connection.open},status=HTTPStatus.OK)


@app.get("/workerhealthcheck")
async def queryworker_healthcheck(request:Request):
    try:
        logger.info(f"Generating worker healthcheck")
        result = {}
        service_name:str
        for service_name in app.ctx.all_slice_services:
            healthcheck_url = f"http://{service_name}:8765/healthcheck"
            async with aiohttp.ClientSession() as session:
                async with session.get(healthcheck_url) as resp:
                    result[service_name] = {
                        "status":resp.status,
                        "response": await resp.text()
                    }
                    
            
        return response.json(result,status=HTTPStatus.OK)
    except Exception:
        logger.error(traceback.format_exc())
        return response.text(body="Failed to fetch health check for one or more workers. Perhaps connections have not been established yet. Run http://\{orchestrator_ip\}:8765/activateworkers",status=HTTPStatus.BAD_REQUEST)


if __name__ == "__main__":
    app.run(host=CURR_HOST,port=8765,access_log=True,debug=True)



from ..configure.configure import Configuration,SliceBlock
from ..backend._api_deploy import PapdlService,PapdlSliceService,PapdlOrchestratorService
from ..backend.common import BenchmarkPreferences,prepare_logger,SplitStrategy
from ..backend.api_common import PapdlAPIContext
from logging import DEBUG
from typing import List
from docker.models.services import Service


def deploy_configuration(configuration:Configuration,debug:bool=False):
    
    logger = prepare_logger()
    
    bp = configuration["benchmark_preferences"]
    input_shape = configuration["input_shape"]
    pac = PapdlAPIContext(preference=bp)

    slice_services:List[PapdlSliceService] = [PapdlSliceService(pac,sb) for sb in configuration["blocks"]]
    
    orchestrator_service = PapdlOrchestratorService(pac,configuration)
    
    orchestrator_service.spawn(forward_service=slice_services[0],slices=slice_services,input_shape=input_shape)
    for i in range(len(slice_services)-1):
        curr:PapdlSliceService = slice_services[i]
        forward:PapdlSliceService = slice_services[i+1]
        curr.spawn(forward_service=forward,debug=debug)
    slice_services[-1].spawn(forward_service=orchestrator_service,debug=debug)
        
    logger.info("Issued service startup request")
    all_services = [*slice_services,orchestrator_service]
    s:Service
    for s in all_services:
        logger.info(s)
    
    



from .api_common import PapdlAPIContext, prepare_build_context, copy_app, AppType,get_docker_logs,get_service_status
from ..configure.configure import Configuration, SliceBlock
from os import path,mkdir
from typing import TypedDict,List,Union,Dict,Tuple
from .common import PapdlException,ContainerBehaviourException

from docker.models.services import Service
from docker.models.nodes import Node
from docker.models.images import Image
from docker.types import NetworkAttachmentConfig,RestartPolicy,EndpointSpec
from time import time,sleep


class PapdlService:
    def __init__(self,context:PapdlAPIContext,build_context:str,target_node:Union[str,Node], apptype:AppType,name:str):
        self.build_context:str = build_context
        if isinstance(target_node,str):
            nodes = [d for d in context.devices if d.id == target_node]
            if len(nodes) != 1:
                raise PapdlException("Mismatch between nodes in deployment configuration and currently available nodes in swarm. Please rerun benchmarking and configuration") 
            self.target_node:Node = nodes[0]
        else:
            self.target_node:Node = target_node
        self.context:PapdlAPIContext = context
        self.apptype:AppType = apptype
        self.image_name = f"localhost:443/{context.project_name}/{name}"
        self.service_name = f"{context.project_name}_{name}"

        self.url:str = None
        self.outwards_url:str = None
        self.service:Service = None
        self.build()

    def service_spawned(self)->bool:
        return self.service != None
        
    def assign_service(self,url:str,outward_url:str,service:Service):
        self.service = service
        self.outwards_url = outward_url
        self.url = url

    def build(self)->"PapdlService":

        buildargs = {
            "local_user": self.context.local_user,
            "local_uid": str(self.context.local_uid)
        }
        self.context.logger.info(
            f"Building {self.image_name} with context {self.build_context}"
        )
        image, build_logs = self.context.client.images.build(
            path=self.build_context,
            tag=self.image_name,
            buildargs=buildargs,
            labels={
                "papdl": "true",
                "project_name": self.context.project_name,
                "type": AppType.SLICE.value
            }
        )
        get_docker_logs(self.context,build_logs)
        self.context.cleanup_target["images"].append(image)
        self.context.logger.info(f"Pushing image {self.image_name} to registry for distribution")
        self.context.client.images.push(self.image_name)
        self.image = image

    @staticmethod
    def _all_match_any_of(_all: List[str], _any: List[str]) -> bool:
        return all([True if x in _any else False for x in _all])
        
    def wait_for_stability(self,timeout_on:List[str]):
        
        self.context.logger.info(f"Spawning service {self.service.name}...")
        
        start_time = time()
        while True:
            if time() - \
                start_time > self.context.preference["startup_timeout"]:
                    raise ContainerBehaviourException(
                        f"Service {self.service.name} timed out trying to spawn...",service=self.service
                    )
            else:
                service_status = get_service_status(service=self.service)
                self.context.logger.debug(
                    f"Service {self.service.name} has status {service_status}"
                )
                if PapdlService._all_match_any_of(service_status,timeout_on):
                    break
                if PapdlService._all_match_any_of(
                    service_status,
                    ["failed","shutdown","rejected","orphaned","remove"]
                ):
                    raise ContainerBehaviourException(f"Service {self.service.name} was/has {service_status}", service=self.service)
                sleep(1)
                
        
        
        
    def spawn(self,forward_service:"PapdlService",extra_args={})->"PapdlService":
        image_name = self.image.tags[0]
        self.context.logger.info(f"Spawning service for image {image_name}")
        nac = NetworkAttachmentConfig(self.context.network.name)
        es = EndpointSpec(mode="vip")
        
        spawning_configuration = {
            "image":image_name,
            "command":f"python3 -m server",
            "constraints":[f"node.id=={self.target_node.id}"],
            "restart_policy":RestartPolicy(condition="none"),
            "labels":{
                "papdl": "true",
                "project_name": self.context.project_name,
                "type":self.apptype.value
            },
            "name":self.service_name,
            "endpoint_spec":es,
            "networks":[nac],
            "env":[f"FORWARD={forward_service.service_name}"],
            **extra_args
        }
        self.service = self.context.client.services.create(**spawning_configuration)
        self.context.cleanup_target["services"].append(self.service)
        self.wait_for_stability(["running"])
        return self


class PapdlSliceService(PapdlService):
    def __init__(self,context:PapdlAPIContext, sb:SliceBlock):
        slice_indices = sb.slice_index
        slice_start = slice_indices[0]
        slice_end = slice_indices[1]
        name = f"slices_{slice_start}_{slice_end}"

        self.slice_block:SliceBlock = sb
        
        context.logger.info(f"Preparing deployment build for slice {slice_indices}")
        build_context = prepare_build_context(context)
        model_path = path.join(build_context,"model")
        mkdir(model_path)
        model = sb.model
        model.save(model_path,overwrite=True)
        copy_app(context,AppType.SLICE,build_context=build_context)
        target_node = sb.device.name

        super().__init__(context=context,build_context=build_context,target_node=target_node,apptype=AppType.SLICE, name=name)
    
    def spawn(self,forward_service:"PapdlService",debug:bool=False)->"PapdlSliceService":
        env = [
            f"DEBUG={int(debug)}",
            f"FORWARD={forward_service.service_name}"
        ]
        return super().spawn(forward_service=forward_service,extra_args={"env":env})
        
class PapdlOrchestratorService(PapdlService):
    def __init__(self,context:PapdlAPIContext, configuration:Configuration):
        name = f"orchestrator"
        context.logger.info(f"Preparing orchestrator build...")
        build_context = prepare_build_context(context)
        copy_app(context=context,app_type=AppType.ORCHESTRATOR,build_context=build_context)
        source_device = configuration["source_device"].name
        super().__init__(context=context,build_context=build_context,target_node=source_device,apptype=AppType.ORCHESTRATOR,name=name)
        
    def spawn(self,forward_service:"PapdlService",slices:List[PapdlSliceService],input_shape=Tuple[int])->"PapdlOrchestratorService":
        es = EndpointSpec(mode="vip", ports={8765:8765})
        env = [
            f"FORWARD={forward_service.service_name}",
            f"SLICES={','.join([s.service_name for s in slices])}",
            f"INPUTDIMS={','.join([str(s) for s in input_shape])}"
        ]
        return super().spawn(forward_service=forward_service,extra_args={"endpoint_spec":es,"env":env})
    
            
            
            
            
            
        
        



from enum import Enum
from typing import List, Dict, TypedDict,Tuple
from ..slice.slice import Slice
from ..backend.api import PapdlAPI, DeploymentStatus
from ..backend.common import ContainerBehaviourException, BenchmarkPreferences, SplitStrategy
from ..backend.api_common import PapdlAPIContext
from docker.models.nodes import Node
from docker.models.services import Service
from docker.errors import DockerException, APIError
from copy import deepcopy

preferences: BenchmarkPreferences


class BenchmarkResult(TypedDict):
    arg_preferences:BenchmarkPreferences
    # papdl_api:PapdlAPI
    result:Dict
    slice_list:List[Slice]

def benchmark_slices(slice_list: List[Slice],arg_preferences: BenchmarkPreferences) -> BenchmarkResult:
    global preferences
    preferences = arg_preferences
    benchmark_results = scission_strategy(slice_list)

    return BenchmarkResult(arg_preferences=arg_preferences,result=benchmark_results,slice_list=[sl.model for sl in slice_list])


def translate_statistics(ds: DeploymentStatus, statistics: Dict) -> Dict:
    ip_to_node_mapping = {v: k for k, v in ds.node_ip_mapping.items()}
    new_statistics = {}
    for device, device_statistics in statistics.items():
        new_statistics[device] = {
            "model_performance": None,
            "network_performance": {},
            "free_memory": deepcopy(device_statistics)["free_memory"]
            }
        new_statistics[device]["model_performance"] = deepcopy(
            device_statistics["model_performance"])
        for ip, network_benchmark in device_statistics["network_performance"].items(
        ):
            new_statistics[device]["network_performance"][ip_to_node_mapping[ip]] = deepcopy(
                network_benchmark)

    return new_statistics


def scission_strategy(slice_list: List[Slice]) -> Dict:
    global preferences
    statistics = {}
    api: PapdlAPI = None
    ds: DeploymentStatus = None
    try:
        pac = PapdlAPIContext(preference=preferences)
        api = PapdlAPI(context=pac)
        ds = api.deploy_benchmarkers(slice_list)
        deployed_services = ds.node_service_mapping
        node: Node
        service: Service
        for node, service in deployed_services.items():
            statistics[node.id] = api.get_service_logs(service)
    except APIError as e:
        preferences['logger'].error(e.response.text)
    except ContainerBehaviourException as e:
        preferences['logger'].error(e.message)
        preferences['logger'].error("=====SERVICE LOGS=====")
        preferences['logger'].error(api.get_service_logs(e.service))
        exit(1)
    except DockerException as e:
        preferences['logger'].error(
            "Docker Exception occured. Have you started the client?")
        preferences['logger'].error(e)
    finally:
        if api is not None:
            api.cleanup()
    return translate_statistics(ds, statistics)





import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import click
from ..backend.common import prepare_logger,BenchmarkPreferences
from ..benchmark.benchmark import benchmark_slices,SplitStrategy,BenchmarkResult
from logging import DEBUG
import traceback
import dill as pickle

@click.command()
@click.argument("sliced_model_path")
@click.option("-n","--project_name",type=str,default="debug")
@click.option("-o","--output",type=str)
@click.option("-i","--service_idle_detection",type=int,default=6000)
@click.option("-s","--startup_timeout",type=int,default=6000)

@click.option("-b","--model_test_batch_size",type=int,default=1)
@click.option("-r","--model_test_number_of_repeats",type=int,default=10)
@click.option("-d","--bandwidth_test_duration_sec",type=int,default=1)
@click.option("-l","--latency_test_count",type=int,default=100)
@click.option("-m","--free_memory_multiplier",type=float,default=0.8)
#TODO: search constraints
def benchmark(
    sliced_model_path:str,
    project_name:str,
    output:str,
    service_idle_detection:int,
    startup_timeout:int,

    model_test_batch_size:int,
    model_test_number_of_repeats:int,
    bandwidth_test_duration_sec:int,
    latency_test_count:int,
    free_memory_multiplier:float
):
    logger = prepare_logger(DEBUG)
    logger.info("Loading Sliced Models...")
    pref = BenchmarkPreferences(
        service_idle_detection=service_idle_detection,
        split_strategy=SplitStrategy.from_str("scission"),
        logger=logger,
        startup_timeout=startup_timeout,

        model_test_batch_size=model_test_batch_size,
        model_test_number_of_repeats=model_test_number_of_repeats,
        bandwidth_test_duration_sec=bandwidth_test_duration_sec,
        latency_test_count=latency_test_count,
        free_memory_multiplier=free_memory_multiplier,

        project_name=project_name
    )
    try:
        sliced_models = None
        
        with open(sliced_model_path,"rb") as f:
            sliced_models = pickle.load(f)

        benchmark_result:BenchmarkResult = benchmark_slices(sliced_models,arg_preferences=pref)
        
        if(output is None):
            output = "benchmark.pickle"
        
        with open(output, "wb") as f:
            pickle.dump(benchmark_result,f,protocol=pickle.HIGHEST_PROTOCOL)
        logger.info(f"Saving benchmark results as '{output}'")

    except Exception:
        logger.error(traceback.format_exc())
        exit(1)
    


from logging import Logger
from typing import NamedTuple, List, Dict, TypedDict, Union, Tuple, MutableSet
from json import loads, dumps
import heapq
import keras
from ..backend.common import PapdlException, BenchmarkPreferences
import logging
import re
from tabulate import tabulate
from tqdm import tqdm
import math
import pickle
import psutil
import gc
from time import sleep


class Layer():
    def __init__(self, name, model_memory_usage, hidden_memory_usage, io_memory_usage):
        self.name = name
        self.model_memory_usage = model_memory_usage
        self.hidden_memory_usage = hidden_memory_usage
        self.io_memory_usage = io_memory_usage

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other: "Layer"):
        if other is None:
            return False
        if not isinstance(other, Layer):
            return False
        return other.name == self.name

    def __str__(self):
        return self.name


class Worker():
    def __init__(self, name, free_memory):
        self.name = name
        self.free_memory = free_memory

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other: "Worker"):
        if other is None:
            return False
        if not isinstance(other, Worker):
            return False
        return other.name == self.name

    def __str__(self):
        return self.name


class SearchConstraints():
    def __init__(self, layer_must_be_in_device: Dict[Layer, Worker], layer_must_not_be_in_device: Dict[Layer, Worker]):
        self.layer_must_be_in_device: Dict[Layer,
                                           Worker] = layer_must_be_in_device
        self.layer_must_not_be_in_device: Dict[Layer,
                                               Worker] = layer_must_not_be_in_device

    @staticmethod
    def parse_match(match: str) -> Dict[str, List[str]]:
        result = {}
        if match is not None:
            conditions = match.split(",")
            for key_val in conditions:
                split_key_val = key_val.split(":")
                key = split_key_val[0].strip()
                val = split_key_val[1].strip()
                result[key] = val
        return result

    def __getitem__(self, item):
        if item == "layer_must_be_in_device":
            return self.layer_must_be_in_device
        if item == "layer_must_not_be_in_device":
            return self.layer_must_not_be_in_device
        raise KeyError

    def __str__(self):
        return f"must={self.layer_must_be_in_device},mustnot={self.layer_must_not_be_in_device}"

    @staticmethod
    def coarsce_type(dict_to_coersce: Dict[str, str]) -> Dict[Layer, Worker]:
        result: Dict[Layer, Worker] = {}
        for layer_str, worker_str in dict_to_coersce.items():
            result[Layer(name=layer_str, model_memory_usage=None,hidden_memory_usage=None,io_memory_usage=None)] = Worker(
                name=worker_str, free_memory=None)
        return result

    @staticmethod
    def parse_from_str(input_str: str):
        try:
            must_str = re.search(r"(?<=must=\{)(\w|\s|:|,)+(?=\})", input_str)
            mustnot_str = re.search(
                r"(?<=mustnot=\{)(\w|\s|:|,)+(?=\})", input_str)

            layer_must_be_in_device: Dict[Layer, Worker] = {}
            if (must_str is not None):
                layer_must_be_in_device_str_dict = SearchConstraints.parse_match(
                    must_str.group())
                layer_must_be_in_device = SearchConstraints.coarsce_type(
                    layer_must_be_in_device_str_dict)

            layer_must_not_be_in_device: Dict[Layer, Worker] = {}
            if (mustnot_str is not None):
                layer_must_not_be_in_device_str_dict = SearchConstraints.parse_match(
                    mustnot_str.group())
                layer_must_not_be_in_device = SearchConstraints.coarsce_type(
                    layer_must_not_be_in_device_str_dict)

            return SearchConstraints(layer_must_be_in_device=layer_must_be_in_device, layer_must_not_be_in_device=layer_must_not_be_in_device)

        except Exception as e:
            raise ValueError


class SliceBlock(NamedTuple):
    layers: List[Layer]
    slice_index: Tuple[int, int]
    device: Worker
    model: keras.models.Model


class Configuration(TypedDict):
    slices: List[Layer]
    blocks: List[SliceBlock]
    devices: List[Worker]
    constraints: SearchConstraints
    source_device: Worker
    input_shape: Tuple[int]
    benchmark_preferences: BenchmarkPreferences
    penalty: float


class ConfigurationPreferences(TypedDict):
    logger: logging.Logger
    search_constraints: SearchConstraints


class Configurer():

    def __init__(self, logger: Logger):
        self.logger = logger

    def tabulated_print(self, configuration: Configuration):
        slice_blocks: List[SliceBlock] = configuration["blocks"]
        # print(slice_blocks)

        table = []
        table.append(["Slice Indices", "Device", "Device Free Memory (MB)", "Model Memory Usage (MB)",
                     "Hidden Layer Memory Usage (MB)", "IO Memory Usage (MB)"])
        sb: SliceBlock
        for sb in slice_blocks:
            table.append(
                [str(sb.slice_index),
                 sb.device.name,
                 sb.device.free_memory / 1_000_000 * configuration["benchmark_preferences"]["free_memory_multiplier"],
                 sum([
                     l.model_memory_usage
                     for l in sb.layers
                 ]) / 1_000_000,
                 sum([
                     l.hidden_memory_usage
                     for l in sb.layers
                 ]) / 1_000_000,
                 sb.layers[0].io_memory_usage / 1_000_000
                 ]
            )
        self.logger.info(
            "\n" + tabulate(table, headers="firstrow", tablefmt="outline"))
        self.logger.info(f"PENALTY: {configuration['penalty']}")

    class DecisionNode():
        def __init__(self,
                     model: Layer = None,
                     device: Worker = None,
                     paths: List["Configurer.Path"] = []):
            self.model: Layer = model
            self.device: Worker = device
            self.paths: List[Configurer.Path] = paths

        def __lt__(self, other):
            return False
        
        def __hash__(self) -> int:
            if self.model is None:
                return hash("NULL" + "-" + self.device.name)
            else:
                return hash(self.model.name + "-" + self.device.name)

        def __eq__(self, other: "Configurer.DecisionNode") -> bool:
            if other is None:
                return False

            if not isinstance(other, Configurer.DecisionNode):
                return False

            if (
                self.model is None and other.model is not None or
                self.model is not None and other.model is None
            ):
                return False

            if self.model is None and other.model is None:
                return self.device == other.device

            return self.model == other.model and self.device == other.device

        def __str_children(self) -> str:
            result = []
            p: Configurer.Path
            for p in self.paths:
                model_name = p["node"].model.name if p["node"].model is not None else "NULL"
                result.append(
                    f"*{model_name}-{p['node'].device.name}-{p['penalty']}*")
            return "\n".join(result)

        def debug_str(self) -> str:
            model_name = "NULL" if self.model is None else self.model.name
            return f"<NODE model:{model_name} device:{self.device.name} children=[\n{self.__str_children()}]>"

        def __str__(self) -> str:
            model_name = "NULL" if self.model is None else self.model.name
            return f"<NODE model:{model_name} device:{self.device.name}>"

    class Path():
        def __init__(self, node: "Configurer.DecisionNode", penalty: float):
            self.node = node
            self.penalty = penalty

        def __eq__(self, other: "Configurer.Path"):
            return isinstance(other, Configurer.Path) and other.penalty == self.penalty

        def __lt__(self, other: "Configurer.Path"):
            return isinstance(other, Configurer.Path) and other.penalty < self.penalty

    class OptimalPath(NamedTuple):
        path: List["Configurer.DecisionNode"]
        penalty: float

    class SearchStatus(NamedTuple):
        total_distance: float
        path: "Configurer.Path"

    def __valid_path(path: List[DecisionNode], constraints: SearchConstraints, benchmark_pref: BenchmarkPreferences):
        node: Configurer.DecisionNode
        for node in path:
            for model, device in constraints["layer_must_be_in_device"].items(
            ):
                if model == node.model and device != node.device:
                    return False

        node: Configurer.DecisionNode
        for node in path:
            for model, device in constraints["layer_must_not_be_in_device"].items():
                if model == node.model and device == node.device:
                    return False

        node: Configurer.DecisionNode
        worker_memory_usage: Dict[Worker, int] = {}
        prevNode = path[0]
        for node in path:
            if node.device not in worker_memory_usage.keys():
                worker_memory_usage[node.device] = 0
            if node.model is None:
                continue
            worker_memory_usage[node.device] += node.model.model_memory_usage + \
                node.model.hidden_memory_usage

            if node.device != prevNode:
                worker_memory_usage[node.device] += node.model.io_memory_usage
                prevNode = node.device

            for worker, memory_usage in worker_memory_usage.items():
                if memory_usage > worker.free_memory * benchmark_pref["free_memory_multiplier"]:
                    return False
        return True

    def __jumps(path: List[DecisionNode]):
        jump_count = 0
        prev_device: Worker = path[0].device
        n: Configurer.DecisionNode
        for n in path[1:]:
            if n.device != prev_device:
                jump_count += 1
                prev_device = n.device

        return jump_count

    def __free_memory_variance(path: List[DecisionNode]):
        node: Configurer.DecisionNode
        worker_memory_usage: Dict[Worker, int] = {}

        prevDevice:Worker = path[0].device
        for node in path:
            if node.device not in worker_memory_usage.keys():
                worker_memory_usage[node.device] = 0
            if node.model is None:
                continue
            worker_memory_usage[node.device] += node.model.model_memory_usage + node.model.hidden_memory_usage
            if(prevDevice != node.device):
                worker_memory_usage[node.device] += node.model.io_memory_usage
                

        average = [memory_usage / worker.free_memory for worker,
                   memory_usage in worker_memory_usage.items()]

        variance = sum([
            (memory_usage / worker.free_memory - avg)
            for avg, (worker, memory_usage)
            in zip(
                average,
                worker_memory_usage.items())
        ]) / len(worker_memory_usage)
        return variance

    def __find_shortest_loop(
        start_node: DecisionNode,
        constraints: SearchConstraints,
        benchmark_pref: BenchmarkPreferences,
        depth: int,
        width: int
    ) -> Union[OptimalPath, None]:
        visited = set()
        queue = [(0, start_node, [], visited)]
        frontier_depth = 1
        with tqdm(total=depth + 2) as pbar:
            while queue:
                penalty, current_node, traversed_nodes, visited = heapq.heappop(
                    queue)
                

                if current_node in visited and current_node == start_node:
                    pbar.update(1)
                    return Configurer.OptimalPath(path=traversed_nodes + [current_node], penalty=penalty)

                if current_node not in visited:
                    visited.add(current_node)
                    traversed_nodes = traversed_nodes + [current_node]

                    for p in current_node.paths:
                        new_penalty = penalty + p.penalty
                        new_path = traversed_nodes + [p.node]


                        if Configurer.__valid_path(new_path, constraints, benchmark_pref) and new_penalty != float('inf'):
                            heapq.heappush(
                                queue, (new_penalty, p.node, traversed_nodes, visited.copy()))
                            
                        else:
                            pass

                        if len(new_path) > frontier_depth:
                            pbar.update(1)
                            frontier_depth = len(new_path)
                        
                        # if psutil.virtual_memory().free >> 30 <= 1:
                        #     queue = queue[:len(queue)//2]
                        #     heapq.heapify(queue)
                        #     gc.collect()
                # print("=====")
        return None

    def __calculate_performance_penalty(
            benchmark_result: Dict,
            destination: Worker,
            model: Layer) -> float:
        return benchmark_result[destination.name]["model_performance"][model.name]["benchmark_time"]

    def __calculate_network_penalty(
        benchmark_result: Dict,
        source: Worker,
        destination: Worker,
        filesize_to_send: int
    ) -> float:
        stats = benchmark_result[source.name]["network_performance"][destination.name]
        latency = stats["latency"]["rtt_avg_ms"] / 1000 # Get in seconds
        bandwidth = stats["bandwidth"]["sent_bps"]
        return (latency + (filesize_to_send / bandwidth))

    def __calculate_network_penalty_from_model(
        benchmark_result: Dict,
        source: Worker,
        destination: Worker,
        model: Layer
    ) -> float:
        filesize_to_send = benchmark_result[source.name]["model_performance"][model.name]["benchmark_size"]
        return Configurer.__calculate_network_penalty(
            benchmark_result=benchmark_result,
            source=source,
            destination=destination,
            filesize_to_send=filesize_to_send
        )

    def __generate_path(
        benchmark_result: Dict,
        source: Worker,
        destination: Worker,
        next_model: Layer,
        input_size: int
    ) -> Path:
        global visited_node_map

        penalty: float = 0
        if next_model is not None:
            penalty = Configurer.__calculate_network_penalty_from_model(
                benchmark_result=benchmark_result,
                source=source,
                destination=destination,
                model=next_model
            )
            penalty += Configurer.__calculate_performance_penalty(
                benchmark_result=benchmark_result,
                destination=destination,
                model=next_model
            )
        else:
            penalty += Configurer.__calculate_network_penalty(
                benchmark_result=benchmark_result,
                source=source,
                destination=destination,
                filesize_to_send=input_size
            )

        temp: Configurer.DecisionNode = Configurer.DecisionNode(
            model=next_model, device=destination)
        path: Configurer.Path
        if temp in visited_node_map:
            path = Configurer.Path(
                node=visited_node_map.get(temp),
                penalty=penalty
            )
        else:
            path = Configurer.Path(
                node=Configurer.DecisionNode(
                    model=next_model,
                    device=destination,
                    paths=[]
                ),
                penalty=penalty
            )
        return path

    def __rec_construct_path(
            benchmark_result: Dict,
            models_left: List[Layer],
            currNode: DecisionNode,
            source_device: Worker,
            devices: List[Worker],
            input_size: int):

        global visited_node_map

        if currNode in visited_node_map:
            return

        visited_node_map[currNode] = currNode
        if len(models_left) == 0:
            currNode.paths = [
                Configurer.__generate_path(
                    benchmark_result=benchmark_result,
                    source=currNode.device,
                    destination=source_device,
                    next_model=None,
                    input_size=input_size
                )
            ]
            return
        else:
            paths = [
                Configurer.__generate_path(
                    benchmark_result=benchmark_result,
                    source=currNode.device,
                    destination=d,
                    next_model=models_left[0],
                    input_size=input_size
                )
                for d in devices
            ]
            currNode.paths = paths
            path: Configurer.Path
            for path in currNode.paths:
                nextNode = path.node
                Configurer.__rec_construct_path(
                    benchmark_result=benchmark_result,
                    models_left=models_left[1:],
                    currNode=nextNode,
                    source_device=source_device,
                    devices=devices,
                    input_size=input_size
                )

    def __fetch_model_from_nodes(nodes: List[DecisionNode], models: List[keras.models.Model]) -> List[keras.models.Model]:
        result: List[keras.models.Model] = []
        # jprint([
        # j    n.model.name
        # j    for n in nodes
        # j    if (n.model is not None)
        # j    else None])
        # print([m.name for m in models])
        for n in nodes:
            search = [
                m for m in models if n.model is not None and m.name == n.model.name]
            if len(search) != 1:
                raise PapdlException(
                    "Model names in benchmark.json does not match model names for the ones used for benchmarking. Rerun benchmarking process...")
            result.append(search[0])
        return result

    def __merge_models(models: List[keras.models.Model]) -> keras.models.Model:
        result_model = keras.models.Sequential()
        for model in models:
            result_model.add(model)
        result_model.build()
        return result_model

    def __generate_blocks(
        op: OptimalPath,
        models: List[keras.models.Model]
    ) -> List[SliceBlock]:
        l = 1
        r = 2
        slices: List[SliceBlock] = []
        while r < len(op.path):
            # print((l,r,slices))
            if op.path[r].device != op.path[l].device:
                nodes_slice = op.path[l:r]
                s = SliceBlock(
                    layers=[
                        n.model
                        for n in nodes_slice
                    ],
                    slice_index=(l-1, r-1),
                    device=op.path[l].device,
                    model=Configurer.__merge_models(
                        Configurer.__fetch_model_from_nodes(
                            nodes_slice, models)
                    )
                )
                slices.append(s)
                l = r
            r += 1

        if len(slices) == 0:
            return [
                SliceBlock(
                    layers=[n.model for n in op.path if n.model is not None],
                    slice_index=(0, len(op.path)-1),
                    device=op.path[1].device,
                    model=Configurer.__merge_models(
                        Configurer.__fetch_model_from_nodes(
                            [n for n in op.path if n.model is not None], models)
                    )
                )
            ]

        return slices

    def parse_from_benchmark(
        self,
        benchmark_result: Dict,
        source_device: Union[Worker, str],
        input_size: int,
        search_constraints: SearchConstraints,
        model_list: List[keras.models.Model],
        benchmark_pref: BenchmarkPreferences
    ) -> Configuration:
        devices: List[Worker] = [
            Worker(k, benchmark_result[k]["free_memory"]) for k in list(
                benchmark_result.keys())]
        
        

        self.logger.debug("Benchmark Result")
        self.logger.debug(benchmark_result)
        sd: Worker = None
        if isinstance(source_device, Worker):
            sd = source_device
        if isinstance(source_device, str):
            sd = Worker(
                name=source_device, free_memory=benchmark_result[source_device]["free_memory"])

        models: List[Layer] = []

        def model_total_ordering(k: str):
            _split = k.split("_")
            if len(_split) == 1:
                return 0
            else:
                return int(_split[1])
        sorted_models = sorted(list(
            benchmark_result[sd.name]["model_performance"].keys()), key=model_total_ordering)
        for m_key in sorted_models:
            model_dict = benchmark_result[sd.name]["model_performance"][m_key]
            models.append(
                Layer(
                    name=m_key,
                    model_memory_usage=model_dict["benchmark_model_memory_usage"],
                    hidden_memory_usage=model_dict["benchmark_hidden_and_input_memory_usage"],
                    io_memory_usage=model_dict["benchmark_input_memory_multiplier"]
                )
            )
        # self.tabulated_print_devices(devices,benchmark_pref)

        global visited_node_map

        visited_node_map = {}

        head = Configurer.DecisionNode(
            model=None,
            device=sd
        )
        Configurer.__rec_construct_path(
            benchmark_result=benchmark_result,
            models_left=models,
            currNode=head,
            source_device=sd,
            devices=devices,
            input_size=input_size
        )

        self.logger.info(
            f"Searching path with benchmark preferences: {benchmark_pref} and search preferences: {search_constraints}")
        depth = len(models)
        width = len(devices)

        shortest_loop = Configurer.__find_shortest_loop(
            start_node=head,
            constraints=search_constraints,
            benchmark_pref=benchmark_pref,
            depth=depth,
            width=width
        )
        
        # [print((n.device.name, n.model.name if n.model is not None else None)) for n in shortest_loop.path]
        

        # shortest_loop = None

        # with open("shortest_loop_cache.pickle","rb") as f:
        #     shortest_loop = pickle.load(f)

        if shortest_loop is None:
            self.logger.error("No path found with the provided constraints")
            exit(1)

        # print([f"model:{n.model} device:{n.device}" for n in shortest_loop.path])
        blocks = Configurer.__generate_blocks(shortest_loop, model_list)

        # print([b.device.name for b in blocks])

        input_shape = blocks[0].model.input_shape[1:]

        config = Configuration(
            slices=models,
            blocks=blocks,
            devices=devices,
            constraints=search_constraints,
            source_device=sd,
            input_shape=input_shape,
            benchmark_preferences=benchmark_pref,
            penalty=shortest_loop.penalty
        )
        return config



import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import click
from .configure import Configuration,SearchConstraints,Configurer
from ..backend.common import prepare_logger
from ..benchmark.benchmark import BenchmarkResult
from logging import DEBUG
import dill as pickle
import traceback


@click.command()
@click.argument("benchmark_result_path")
@click.argument("source_device")
@click.argument("input_size")
@click.option("-o","--output")
@click.option("-c","--search_constraints",type=str,default="")
def configure(
    benchmark_result_path:str,
    source_device:str,
    input_size:str,
    output:str,
    search_constraints:str
):
    try:
        logger = prepare_logger(DEBUG)
        configurer = Configurer(logger=logger)
        
        benchmark_result:BenchmarkResult = None
        with open(benchmark_result_path,"rb") as f:
            benchmark_result = pickle.load(f)
        
        # sc = SearchConstraints(layer_must_be_in_device={},layer_must_not_be_in_device={})
        sc:SearchConstraints = SearchConstraints.parse_from_str(search_constraints)
        
        configuration:Configuration =  configurer.parse_from_benchmark(
            benchmark_result=benchmark_result["result"],
            source_device=source_device,
            input_size=int(input_size), 
            search_constraints=sc,
            model_list=benchmark_result["slice_list"],
            benchmark_pref=benchmark_result["arg_preferences"])
    
        if output is None:
            output = "configuration.pickle"
        with open(output, "wb") as f:
            pickle.dump(configuration,f,protocol=pickle.HIGHEST_PROTOCOL)
        logger.info(f"Saving sliced model list as '{output}'")
        configurer.tabulated_print(configuration=configuration)
    except Exception as e:
        logger.error(traceback.format_exc())
    
    
    



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['papdl',
 'papdl.backend',
 'papdl.backend.containers.Benchmarker.app',
 'papdl.backend.containers.Orchestrator.app',
 'papdl.backend.containers.Orchestrator.app.proto',
 'papdl.backend.containers.Slice.app',
 'papdl.backend.containers.Slice.app.proto',
 'papdl.benchmark',
 'papdl.clean',
 'papdl.configure',
 'papdl.deploy',
 'papdl.slice']

package_data = \
{'': ['*'],
 'papdl.backend': ['certificates/.stub',
                   'containers/Benchmarker/*',
                   'containers/Orchestrator/*',
                   'containers/Slice/*',
                   'registry_volume/*']}

install_requires = \
['aiohttp>=3.8.4,<4.0.0',
 'asynchttp>=0.0.4,<0.0.5',
 'asyncio>=3.4.3,<4.0.0',
 'asyncstdlib>=3.10.5,<4.0.0',
 'autopep8>=2.0.1,<3.0.0',
 'click>=8.1.3,<9.0.0',
 'colorama>=0.4.6,<0.5.0',
 'coloredlogs>=15.0.1,<16.0.0',
 'crossplane>=0.5.8,<0.6.0',
 'dill>=0.3.6,<0.4.0',
 'docker>=6.0.1,<7.0.0',
 'filprofiler>=2023.3.0,<2024.0.0',
 'iperf3>=0.1.11,<0.2.0',
 'ipykernel>=6.21.3,<7.0.0',
 'jsonpickle>=3.0.1,<4.0.0',
 'keras>=2.11.0,<3.0.0',
 'lz4>=4.3.2,<5.0.0',
 'matplotlib>=3.7.1,<4.0.0',
 'memory-profiler>=0.61.0,<0.62.0',
 'notebook>=6.5.3,<7.0.0',
 'numpy>=1.24.1,<2.0.0',
 'pandas>=1.5.2,<2.0.0',
 'psutil>=5.9.4,<6.0.0',
 'pympler>=1.0.1,<2.0.0',
 'pyopenssl>=23.0.0,<24.0.0',
 'pythonping>=1.1.4,<2.0.0',
 'pyyaml>=6.0,<7.0',
 'random-word>=1.0.11,<2.0.0',
 'sanic>=22.12.0,<23.0.0',
 'tabulate>=0.9.0,<0.10.0',
 'tensorflow>=2.11.0,<3.0.0',
 'termcolor>=2.1.1,<3.0.0',
 'tqdm>=4.64.1,<5.0.0',
 'uproot>=5.0.4,<6.0.0',
 'websockets>=10.4,<11.0']

entry_points = \
{'console_scripts': ['papdl = papdl.cli:main']}

setup_kwargs = {
    'name': 'papdl',
    'version': '0.1.2',
    'description': '',
    'long_description': '\n# PAPDL - Making Distributed Model Deployment Easier\n\nDistributed systems and edge computing have become increasingly important due to the rapid growth of data generated and the demand for scalable computer systems. In parallel to this growth, Deep Neural Networks have shown exceptional results in a variety of tasks to be performed on the data from image recognition to video processing. PAPDL is a command line program that addresses the challenges of Deep Neural Network (DNN) predictions on real-time streaming data. The proposed program aims to decouple DNN layers from the original model, encapsulate them using containerized services, and deploy them across of set of distributed systems. The decoupling operation takes resource utilization metrics for executing a DNN layer into account in order to prevent certain devices from being overloaded and also aims to minimize the latency for getting a prediction for a given input. PAPDL can orchestrate services running across the distributed system to perform prediction that is semantically equivalent to running the entire model as a monolithic architecture on a single device.\n\n## Requirements and Installation\n\n1. Install Docker https://docs.docker.com/get-docker/\n2. Install PAPDL `pip3 install papdl`\n\n\n## Usage\n\n### 1. Prepare a set of devices which can communicate with each other within a local network or over the internet.\n\nThis example will show a practical use case of optimally deploying a VGG16 Image classification model across a Docker Swarm. We will be using the following devices to run the model:\n\n| Device Name | CPU Cores | Arch    | Memory | Swap |\n|-------------|-----------|---------|--------|------|\n| vm4         | 1         | x86\\_64 | 2GB    | 2GB  |\n| vm5         | 2         | x86\\_64 | 4GB    | 4GB  |\n| pc5         | 4         | x86\\_64 | 32GB   | 8GB  |\n\n\n\n### 2. Setup Docker Swarm configuration.\n\n\n```bash\n$ ssh pc5\n$ docker swarm init\ndocker swarm join --token {JOIN_TOKEN} {IP ADDRESS}\n```\n\nOn one of the devices, run the following command to initialize a Docker Swarm and the manager node. For the purposes of the example, pc5 is used.\n\n\n### 3. Save any arbitrary deep neural network model on the manager node device.\n\n```python\n\nimport keras.applications as apps\nfrom keras.models import save_model\n\nsave_model(apps.VGG16(), filepath=f"vgg16_model")\n```\n\n### 4. Slice the models into layers with the papdl slice command.\n\n```bash\npapdl slice vgg16_model -o sliced_vgg16_model\n```\n\n### 5. Compile layer performance statistics and inter-device network communications statistics with the papdl benchmark command.\n\n```bash\npapdl benchmark sliced_vgg16_model -o benchmarked_vgg16_model\n```\n\n### 6. Determine the node ID of the source device from which input data will be provided from\n\n\n```bash\n$ docker node ls\n \nVM4 {VM4_NODE_ID}\nVM5 {VM5_NODE_ID}\nPC5 {PC5_NODE_ID}\n\n```\n\nFor this guide, all inputs will be provided from vm4 which has the strongest memory and compute constraints.\n\n### 7. Generate optimal deployment configuration based on aggregated data.\n\n```bash\npapdl configure benchmarked_vgg16_model {VM4_NODE_ID} 150528 -o configured_vgg16_model\n```\n\nThe configure step takes two additional arguments {SOURCE_DEVICE} and {INPUT_SIZE}. {SOURCE_DEVICE} indicates where the source of inputs to perform predictions are coming from.  The {INPUT_SIZE} can be calculated multiplying the input shape of VGG16 which is `224 * 224 * 3 * {batch_size}`. The batch size for this guide will be `1` resulting in an input shape of `224 * 224 * 3 * 1 = 150528`.\n\n\n### 8. Deploy the model.\n\n```bash\npapdl deploy configured_vgg16_model\n```\n\nBased on the optimal deployment configuration, we can now containerize sets of layers and distribute them across the configured Docker swarm as micro services. We can verify the instantiation of the services from the following command:\n\n```bash\n$ docker service ls\n```\n\n### 9. Activate a chain of connections between all the containerized micro services:\n\n```bash\n$ ssh vm4\n$ curl localhost:8765/connect\n$ curl localhost:8765/activate\n```\n\n### 10. Perform prediction!\n\nNow we are ready to perform prediction! The `get_papdl_prediction` method shows how we make predictions with PAPDL and `get_local_prediction` method shows how we traditionally make predictions on input using a monolithic DNN architecture running on a single device.\n\n```python\n"""\nLOCAL MONOLITHIC PREDICTION\n"""\ndef get_local_prediction(image):\n    output = model.predict(input)\n    prede = vgg16.decode_predictions(output, top=1)\n    return pred\n\n"""\nDISTRIBUTED PREDICTION WITH PAPDL\n"""\n\ndef get_papdl_prediction(input):\n    buff = io.BytesIO()\n    np.save(buff,input)\n    buff.seek(0)\n\n    response = requests.post("http://localhost:8765/input",data=buff.read(), verify=False, stream=True)\n\n    result_buff = io.BytesIO()\n    result_buff.write(response.content)\n    result_buff.seek(0)\n    result_np = np.load(result_buff)\n\n    pred = vgg16.decode_predictions(result_np,top=1)\n    return pred\n\n\n\ndef predict(dataset):\n    fix,ax = plt.subplots(nrows=1,ncols=len(dataset),figsize=(20,20))\n    for i,(image,sample) in enumerate(dataset):\n        # display(image)\n        input = sample.reshape((-1,224,224,3))\n        pred = get_papdl_prediction(input)\n\n        ax[i].imshow(sample.astype("uint8"),interpolation="nearest")\n        ax[i].get_xaxis().set_visible(False)\n        ax[i].get_yaxis().set_visible(False)\n        ax[i].set_title(f"{pred[0][0][2]} {pred[0][0][1]}")\n\ndef load_data(images):\n    result = []\n    for filename in images:\n        image = load_img(filename,target_size=(224,224))\n        result.append((image, img_to_array(image)))\n    return result\n\npredic(dataset)\n```\n\nIf all goes well, we can verify the output below:\n\n![PAPDL prediction](./scripts/predictions/outputs/papdl.png)\n\n\n## That easy?\n\nYes!\n\n',
    'author': 'Tamim Azmain',
    'author_email': 'maat1@st-andrews.ac.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)




import sys
from time import sleep
import os
import json
import lz4.frame


from websockets.client import connect as ws_connect
from websockets.server import serve as ws_serve
from websockets.datastructures import Headers,HeadersLike
from http import HTTPStatus

import numpy as np
import io
from typing import Optional,Dict,Tuple
import logging
import asyncio
from keras.models import load_model
from keras import Model
import getpass
import traceback
import uproot

os.environ["TF_CPP_MIN_LOG_LEVEL"] = '3'

def isDebug()->bool:
    return int(os.environ.get("DEBUG")) == 1


def prepare_logger()->logging.Logger:
    logger = logging.getLogger("server")
    server_logger_handler: logging.StreamHandler = logging.StreamHandler()
    server_logger_handler.setFormatter(
        logging.Formatter("%(asctime)s [%(name)-6s] %(levelname)-8s %(message)s"
                            ))
    logger.addHandler(server_logger_handler)
    logger.setLevel(logging.DEBUG)

    if isDebug():
        logging.getLogger("websockets").addHandler(logging.NullHandler())
        logging.getLogger('asyncio').setLevel(logging.ERROR)
        logging.getLogger('asyncio.coroutines').setLevel(logging.ERROR)
        logging.getLogger('websockets.server').setLevel(logging.ERROR)
        logging.getLogger('websockets.protocol').setLevel(logging.ERROR)
        logger.propagate = False
    return logger
    
def get_next_url()->str:
    forward_service_name = os.environ.get("FORWARD")
    return f"ws://{forward_service_name}:8765"

    
logger = prepare_logger()
forward_connection = None
forward_url = get_next_url()
model:Model = load_model(f"/home/{getpass.getuser()}/model")
model.compile()

CURR_HOST = "0.0.0.0"
CURR_HOST_PORT =  8765

print("UPROOT SERIALIZATION")

async def forward(websocket):
    async for data in websocket:
        try:
            input_buff = io.BytesIO()
            input_buff.write(lz4.frame.decompress(data))
            input_buff.seek(0)

            requestId = input_buff.read(36).decode("utf-8")
            model_output = model.predict(np.load(input_buff))

            output_buff = io.BytesIO()
            output_buff.write(bytes(requestId,"utf-8"))
            np.save(output_buff, model_output)
            output_buff.seek(0)

            print("SENDING!",flush=True)
            await forward_connection.send(lz4.frame.compress(output_buff.read()))
            print("SENT!",flush=True)
        except:
            logger.error("Caught Exception! Dropping input...")
            logger.error(traceback.format_exc())

async def process_request(path:str,request_headers:Headers)->Optional[Tuple[HTTPStatus,HeadersLike,bytes]]:
    global forward_connection
    global forward_url
    
    # Continue with prediction
    logger.info(f"path={path}")
    logger.info(f"request_headers={request_headers}")

    try:
        if path == "/predict":
            return None
        
        if path == "/connect":
            logger.info(f"Attempting to connect to forward_url: {forward_url}/predict")
            fwu = f"{forward_url}/predict"
            forward_connection = await ws_connect(fwu,open_timeout=None,max_size=sys.maxsize,read_limit=sys.maxsize,write_limit=sys.maxsize)
            logger.info(f"Successfully connected to forward_url: {fwu}")
            return (HTTPStatus.OK,{},b"")
        
        if path == "/healthcheck":
            logger.info("Generating healthcheck")
            return (HTTPStatus.OK, {}, bytes(json.dumps({
                "forward_url":forward_url,
                "connected": forward_connection is not None and forward_connection.open,
                "model_name":model.name
            }),"utf-8"))
    except Exception as e:
        fe = traceback.format_exc()
        logger.error(fe)
        return (HTTPStatus.BAD_REQUEST,{},bytes(str(fe),"utf-8"))

async def serve():
    logger.info("Attempting to serve current websocket server...")
    configuration = {
        "ws_handler":forward,
        "host":CURR_HOST,
        "port":CURR_HOST_PORT,
        "process_request":process_request,
        "read_limit":sys.maxsize,
        "max_size":sys.maxsize,
        "write_limit":sys.maxsize
    }
    if isDebug():
        configuration["logger"] = logger
    async with  ws_serve(**configuration):
        await asyncio.Future()


if __name__ == "__main__":
    asyncio.run(serve())



import numpy as np
from typing import Callable
from websockets.client import connect as ws_connect
from websockets.server import serve as ws_serve
from time import sleep
import io


class Wsc():
    def __init__(self, id: str, host: str, port: int, logger):
        self.id = id
        self.host = host
        self.port = port
        self.conn = None
        self.connected: bool = False
        self.url = f"ws://{self.host}:{self.port}"
        self.server_logger = logger

    def serve(self, handler: Callable):
        if not self.connected:
            self.server_logger.info(f"Attempting to serve {self.url}")
            self.conn = ws_serve(handler, self.host, self.port)
            self.connected = True
            self.server_logger.info(f"Served {self.url}")

    async def connect(self):
        while not self.connected:
            try:
                self.server_logger.info(f"Attempting forward_out {self.url}")
                self.conn = await ws_connect(self.url)
                self.connected = True
                self.server_logger.info(f"Connected to {self.url}")
            except OSError as e:
                self.server_logger.info(e)
                self.server_logger.info(f"Reattempting forward_out {self.url}")
                sleep(1)

    async def send(self, array: np.ndarray):
        write_buff = io.BytesIO()
        np.save(write_buff, array, allow_pickle=True)
        write_buff.seek(0)
        await self.conn.send(write_buff)

    def read_np(data) -> np.ndarray:
        read_buff = io.BytesIO()
        read_buff.write(data)
        read_buff.seek(0)
        return np.load(read_buff, allow_pickle=True)

    def close(self):
        self.conn.close()
